# Copyright 2022 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This is the top-most `__init__.py` file of MrMustard package."""

from rich import print

from ._version import __version__
from .utils.settings import *
from .utils.filters import add_complex_warning_filter


def version():
    r"""Version number of Mr Mustard.

    Returns:
      str: package version number
    """
    return __version__


def about():
    """Mr Mustard information.

    Prints the installed version numbers for Mr Mustard and its dependencies,
    and some system info. Please include this information in bug reports.

    **Example:**

    .. code-block:: pycon

        >>> mm.about()
        Mr Mustard: a differentiable bridge between phase space and Fock space.
        Copyright 2021 Xanadu Quantum Technologies Inc.

        Python version:            3.6.10
        Platform info:             Linux-5.8.18-1-MANJARO-x86_64-with-arch-Manjaro-Linux
        Installation path:         /home/mrmustard/
        Mr Mustard version:        0.1.0
        Numpy version:             1.21.4
        Numba version:             0.48.0
        Scipy version:             1.7.3
        The Walrus version:        0.17.0
        TensorFlow version:        2.7.0
        Torch version:             1.10.0+cu102
    """
    # pylint: disable=import-outside-toplevel
    import os
    import platform
    import sys

    import numba
    import numpy
    import scipy
    import tensorflow
    import thewalrus

    # a QuTiP-style infobox
    print("\nMr Mustard: a differentiable bridge between phase space and Fock space.")
    print("Copyright 2021 Xanadu Quantum Technologies Inc.\n")

    print("Python version:            {}.{}.{}".format(*sys.version_info[0:3]))
    print("Platform info:             {}".format(platform.platform()))
    print("Installation path:         {}".format(os.path.dirname(__file__)))
    print("Mr Mustard version:        {}".format(__version__))
    print("Numpy version:             {}".format(numpy.__version__))
    print("Numba version:             {}".format(numba.__version__))
    print("Scipy version:             {}".format(scipy.__version__))
    print("The Walrus version:        {}".format(thewalrus.__version__))
    print("TensorFlow version:        {}".format(tensorflow.__version__))


# filter tensorflow cast warnings
add_complex_warning_filter()



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BackendBase:
    r"""
    A base class for backends.
    """

    def __init__(self, name):
        self._name = name

    @property
    def name(self):
        r"""
        The name of this backend.
        """
        return self._name



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=redefined-outer-name

"""
This module contains functions for performing calculations on Fock states.
"""

from functools import lru_cache
from typing import List, Optional, Sequence, Tuple

import numpy as np

from mrmustard import math, settings
from mrmustard.math.lattice import strategies
from mrmustard.math.caching import tensor_int_cache
from mrmustard.math.tensor_wrappers.mmtensor import MMTensor
from mrmustard.physics.bargmann import (
    wigner_to_bargmann_Choi,
    wigner_to_bargmann_psi,
    wigner_to_bargmann_rho,
    wigner_to_bargmann_U,
)
from mrmustard.utils.typing import ComplexTensor, Matrix, Scalar, Tensor, Vector

SQRT = np.sqrt(np.arange(1e6))

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~ static functions ~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def fock_state(n: Sequence[int]) -> Tensor:
    r"""Returns a pure or mixed Fock state.

    Args:
        n: a list of photon numbers

    Returns:
        the Fock state up to cutoffs ``n+1``
    """
    psi = np.zeros(np.array(n) + np.ones_like(n), dtype=np.complex128)
    psi[tuple(np.atleast_1d(n))] = 1
    return psi


def autocutoffs(cov: Matrix, means: Vector, probability: float):
    r"""Returns the cutoffs of a Gaussian state by computing the 1-mode marginals until
    the probability of the marginal is less than ``probability``.

    Args:
        cov: the covariance matrix
        means: the means vector
        probability: the cutoff probability

    Returns:
        Tuple[int, ...]: the suggested cutoffs
    """
    M = len(means) // 2
    cutoffs = []
    for i in range(M):
        cov_i = np.array([[cov[i, i], cov[i, i + M]], [cov[i + M, i], cov[i + M, i + M]]])
        means_i = np.array([means[i], means[i + M]])
        # apply 1-d recursion until probability is less than 0.99
        A, B, C = [math.asnumpy(x) for x in wigner_to_bargmann_rho(cov_i, means_i)]
        diag = math.hermite_renormalized_diagonal(A, B, C, cutoffs=[100])
        # find at what index in the cumsum the probability is more than 0.99
        for i, val in enumerate(np.cumsum(diag)):
            if val > probability:
                cutoffs.append(max(i + 1, settings.AUTOCUTOFF_MIN_CUTOFF))
                break
        else:
            cutoffs.append(settings.AUTOCUTOFF_MAX_CUTOFF)
    return cutoffs


def wigner_to_fock_state(
    cov: Matrix,
    means: Vector,
    shape: Sequence[int],
    max_prob: float = 1.0,
    max_photons: Optional[int] = None,
    return_dm: bool = True,
) -> Tensor:
    r"""Returns the Fock representation of a Gaussian state.
    Use with caution: if the cov matrix is that of a mixed state,
    setting return_dm to False will produce nonsense.
    If return_dm=False, we can apply max_prob and max_photons to stop the
    computation of the Fock representation early, when those conditions are met.

    * If the state is pure it can return the state vector (ket) or the density matrix.
        The index ordering is going to be [i's] in ket_i
    * If the state is mixed it can return the density matrix.
        The index order is going to be [i's,j's] in dm_ij

    Args:
        cov: the Wigner covariance matrix
        means: the Wigner means vector
        shape: the shape of the tensor
        max_prob: the maximum probability of a the state (applies only if the ket is returned)
        max_photons: the maximum number of photons in the state (applies only if the ket is returned)
        return_dm: whether to return the density matrix (otherwise it returns the ket)

    Returns:
        Tensor: the fock representation
    """
    if return_dm:
        A, B, C = wigner_to_bargmann_rho(cov, means)
        # NOTE: change the order of the index in AB
        Xmat = math.Xmat(A.shape[-1] // 2)
        A = math.matmul(math.matmul(Xmat, A), Xmat)
        B = math.matvec(Xmat, B)
        return math.hermite_renormalized(A, B, C, shape=shape)
    else:  # here we can apply max prob and max photons
        A, B, C = wigner_to_bargmann_psi(cov, means)
        if max_photons is None:
            max_photons = sum(shape) - len(shape)
        if max_prob < 1.0 or max_photons < sum(shape) - len(shape):
            return math.hermite_renormalized_binomial(
                A, B, C, shape=shape, max_l2=max_prob, global_cutoff=max_photons + 1
            )
        return math.hermite_renormalized(A, B, C, shape=tuple(shape))


def wigner_to_fock_U(X, d, shape):
    r"""Returns the Fock representation of a Gaussian unitary transformation.
    The index order is out_l, in_l, where in_l is to be contracted with the indices of a ket,
    or with the left indices of a density matrix.

    Arguments:
        X: the X matrix
        d: the d vector
        shape: the shape of the tensor

    Returns:
        Tensor: the fock representation of the unitary transformation
    """
    A, B, C = wigner_to_bargmann_U(X, d)
    return math.hermite_renormalized(A, B, C, shape=tuple(shape))


def wigner_to_fock_Choi(X, Y, d, shape):
    r"""Returns the Fock representation of a Gaussian Choi matrix.
    The order of choi indices is :math:`[\mathrm{out}_l, \mathrm{in}_l, \mathrm{out}_r, \mathrm{in}_r]`
    where :math:`\mathrm{in}_l` and :math:`\mathrm{in}_r` are to be contracted with the left and right indices of a density matrix.

    Arguments:
        X: the X matrix
        Y: the Y matrix
        d: the d vector
        shape: the shape of the tensor

    Returns:
        Tensor: the fock representation of the Choi matrix
    """
    A, B, C = wigner_to_bargmann_Choi(X, Y, d)
    # NOTE: change the order of the index in AB
    Xmat = math.Xmat(A.shape[-1] // 2)
    A = math.matmul(math.matmul(Xmat, A), Xmat)
    N = B.shape[-1] // 2
    B = math.concat([B[N:], B[:N]], axis=-1)
    return math.hermite_renormalized(A, B, C, shape=tuple(shape))


def ket_to_dm(ket: Tensor) -> Tensor:
    r"""Maps a ket to a density matrix.

    Args:
        ket: the ket

    Returns:
        Tensor: the density matrix
    """
    return math.outer(ket, math.conj(ket))


def dm_to_ket(dm: Tensor) -> Tensor:
    r"""Maps a density matrix to a ket if the state is pure.

    If the state is pure :math:`\hat \rho= |\psi\rangle\langle \psi|` then the
    ket is the eigenvector of :math:`\rho` corresponding to the eigenvalue 1.

    Args:
        dm (Tensor): the density matrix

    Returns:
        Tensor: the ket

    Raises:
        ValueError: if ket for mixed states cannot be calculated
    """

    is_pure_dm = np.isclose(purity(dm), 1.0, atol=1e-6)
    if not is_pure_dm:
        raise ValueError("Cannot calculate ket for mixed states.")

    cutoffs = dm.shape[: len(dm.shape) // 2]
    d = int(np.prod(cutoffs))
    dm = math.reshape(dm, (d, d))

    eigvals, eigvecs = math.eigh(dm)
    # eigenvalues and related eigenvectors are sorted in non-decreasing order,
    # meaning the associated eigvec to largest eigval is stored last.
    ket = eigvecs[:, -1] * math.sqrt(eigvals[-1])
    ket = math.reshape(ket, cutoffs)

    return ket


def ket_to_probs(ket: Tensor) -> Tensor:
    r"""Maps a ket to probabilities.

    Args:
        ket: the ket

    Returns:
        Tensor: the probabilities vector
    """
    return math.abs(ket) ** 2


def dm_to_probs(dm: Tensor) -> Tensor:
    r"""Extracts the diagonals of a density matrix.

    Args:
        dm: the density matrix

    Returns:
        Tensor: the probabilities vector
    """
    return math.all_diagonals(dm, real=True)


def U_to_choi(U: Tensor, Udual: Optional[Tensor] = None) -> Tensor:
    r"""Converts a unitary transformation to a Choi tensor.

    Args:
        U: the unitary transformation
        Udual: the dual unitary transformation (optional, will use conj U if not provided)

    Returns:
        Tensor: the Choi tensor. The index order is going to be :math:`[\mathrm{out}_l, \mathrm{in}_l, \mathrm{out}_r, \mathrm{in}_r]`
        where :math:`\mathrm{in}_l` and :math:`\mathrm{in}_r` are to be contracted with the left and right indices of the density matrix.
    """
    return math.outer(U, math.conj(U) if Udual is None else Udual)


def fidelity(state_a, state_b, a_ket: bool, b_ket: bool) -> Scalar:
    r"""Computes the fidelity between two states in Fock representation."""
    if a_ket and b_ket:
        min_cutoffs = [slice(min(a, b)) for a, b in zip(state_a.shape, state_b.shape)]
        state_a = state_a[tuple(min_cutoffs)]
        state_b = state_b[tuple(min_cutoffs)]
        return math.abs(math.sum(math.conj(state_a) * state_b)) ** 2

    if a_ket:
        min_cutoffs = [
            slice(min(a, b))
            for a, b in zip(state_a.shape, state_b.shape[: len(state_b.shape) // 2])
        ]
        state_a = state_a[tuple(min_cutoffs)]
        state_b = state_b[tuple(min_cutoffs * 2)]
        a = math.reshape(state_a, -1)
        return math.real(
            math.sum(math.conj(a) * math.matvec(math.reshape(state_b, (len(a), len(a))), a))
        )

    if b_ket:
        min_cutoffs = [
            slice(min(a, b))
            for a, b in zip(state_a.shape[: len(state_a.shape) // 2], state_b.shape)
        ]
        state_a = state_a[tuple(min_cutoffs * 2)]
        state_b = state_b[tuple(min_cutoffs)]
        b = math.reshape(state_b, -1)
        return math.real(
            math.sum(math.conj(b) * math.matvec(math.reshape(state_a, (len(b), len(b))), b))
        )

    # mixed state
    # Richard Jozsa (1994) Fidelity for Mixed Quantum States, Journal of Modern Optics, 41:12, 2315-2323, DOI: 10.1080/09500349414552171

    # trim states to have same cutoff
    min_cutoffs = [
        slice(min(a, b))
        for a, b in zip(
            state_a.shape[: len(state_a.shape) // 2],
            state_b.shape[: len(state_b.shape) // 2],
        )
    ]
    state_a = state_a[tuple(min_cutoffs * 2)]
    state_b = state_b[tuple(min_cutoffs * 2)]
    return math.abs(
        (
            math.trace(
                math.sqrtm(
                    math.matmul(math.matmul(math.sqrtm(state_a), state_b), math.sqrtm(state_a))
                )
            )
            ** 2
        )
    )


def number_means(tensor, is_dm: bool):
    r"""Returns the mean of the number operator in each mode."""
    probs = math.all_diagonals(tensor, real=True) if is_dm else math.abs(tensor) ** 2
    modes = list(range(len(probs.shape)))
    # print("aa", [modes[:k] + modes[k + 1 :] for k in range(len(modes))])
    marginals = [math.sum(probs, axes=modes[:k] + modes[k + 1 :]) for k in range(len(modes))]
    return math.astensor(
        [
            math.sum(marginal * math.arange(len(marginal), dtype=math.float64))
            for marginal in marginals
        ]
    )


def number_variances(tensor, is_dm: bool):
    r"""Returns the variance of the number operator in each mode."""
    probs = math.all_diagonals(tensor, real=True) if is_dm else math.abs(tensor) ** 2
    modes = list(range(len(probs.shape)))
    marginals = [math.sum(probs, axes=modes[:k] + modes[k + 1 :]) for k in range(len(modes))]
    return math.astensor(
        [
            (
                math.sum(marginal * math.arange(marginal.shape[0], dtype=marginal.dtype) ** 2)
                - math.sum(marginal * math.arange(marginal.shape[0], dtype=marginal.dtype)) ** 2
            )
            for marginal in marginals
        ]
    )


def purity(dm: Tensor) -> Scalar:
    r"""Returns the purity of a density matrix."""
    cutoffs = dm.shape[: len(dm.shape) // 2]
    d = int(np.prod(cutoffs))  # combined cutoffs in all modes
    dm = math.reshape(dm, (d, d))
    dm = dm / math.trace(dm)  # assumes all nonzero values are included in the density matrix
    return math.abs(math.sum(math.transpose(dm) * dm))  # tr(rho^2)


def validate_contraction_indices(in_idx, out_idx, M, name):
    r"""Validates the indices used for the contraction of a tensor."""
    if len(set(in_idx)) != len(in_idx):
        raise ValueError(f"{name}_in_idx should not contain repeated indices.")
    if len(set(out_idx)) != len(out_idx):
        raise ValueError(f"{name}_out_idx should not contain repeated indices.")
    if not set(range(M)).intersection(out_idx).issubset(set(in_idx)):
        wrong_indices = set(range(M)).intersection(out_idx) - set(in_idx)
        raise ValueError(
            f"Indices {wrong_indices} in {name}_out_idx are trying to replace uncontracted indices."
        )


def apply_kraus_to_ket(kraus, ket, kraus_in_modes, kraus_out_modes=None):
    r"""Applies a kraus operator to a ket.
    It assumes that the ket is indexed as left_1, ..., left_n.

    The kraus op has indices that contract with the ket (kraus_in_modes) and indices that are left over (kraus_out_modes).
    The final index order will be sorted (note that an index appearing in both kraus_in_modes and kraus_out_modes will replace the original index).

    Args:
        kraus (array): the kraus operator to be applied
        ket (array): the ket to which the operator is applied
        kraus_in_modes (list of ints): the indices (counting from 0) of the kraus operator that contract with the ket
        kraus_out_modes (list of ints): the indices (counting from 0) of the kraus operator that are leftover

    Returns:
        array: the resulting ket with indices as kraus_out_modes + uncontracted ket indices
    """
    if kraus_out_modes is None:
        kraus_out_modes = kraus_in_modes

    if not set(kraus_in_modes).issubset(range(ket.ndim)):
        raise ValueError("kraus_in_modes should be a subset of the ket indices.")

    # check that there are no repeated indices in kraus_in_modes and kraus_out_modes (separately)
    validate_contraction_indices(kraus_in_modes, kraus_out_modes, ket.ndim, "kraus")

    ket = MMTensor(ket, axis_labels=[f"in_left_{i}" for i in range(ket.ndim)])
    kraus = MMTensor(
        kraus,
        axis_labels=[f"out_left_{i}" for i in kraus_out_modes]
        + [f"in_left_{i}" for i in kraus_in_modes],
    )

    # contract the operator with the ket.
    # now the leftover indices are in the order kraus_out_modes + uncontracted ket indices
    kraus_ket = kraus @ ket

    # sort kraus_ket.axis_labels by the int at the end of each label.
    # Each label is guaranteed to have a unique int at the end.
    new_axis_labels = sorted(kraus_ket.axis_labels, key=lambda x: int(x.split("_")[-1]))

    return kraus_ket.transpose(new_axis_labels).tensor


def apply_kraus_to_dm(kraus, dm, kraus_in_modes, kraus_out_modes=None):
    r"""Applies a kraus operator to a density matrix.
    It assumes that the density matrix is indexed as left_1, ..., left_n, right_1, ..., right_n.

    The kraus operator has indices that contract with the density matrix (kraus_in_modes) and indices that are leftover (kraus_out_modes).
    `kraus` will contract from the left and from the right with the density matrix. For right contraction the kraus op is conjugated.

    Args:
        kraus (array): the operator to be applied
        dm (array): the density matrix to which the operator is applied
        kraus_in_modes (list of ints): the indices (counting from 0) of the kraus operator that contract with the density matrix
        kraus_out_modes (list of ints): the indices (counting from 0) of the kraus operator that are leftover (default None, in which case kraus_out_modes = kraus_in_modes)

    Returns:
        array: the resulting density matrix
    """
    if kraus_out_modes is None:
        kraus_out_modes = kraus_in_modes

    if not set(kraus_in_modes).issubset(range(dm.ndim // 2)):
        raise ValueError("kraus_in_modes should be a subset of the density matrix indices.")

    # check that there are no repeated indices in kraus_in_modes and kraus_out_modes (separately)
    validate_contraction_indices(kraus_in_modes, kraus_out_modes, dm.ndim // 2, "kraus")

    dm = MMTensor(
        dm,
        axis_labels=[f"left_{i}" for i in range(dm.ndim // 2)]
        + [f"right_{i}" for i in range(dm.ndim // 2)],
    )
    kraus = MMTensor(
        kraus,
        axis_labels=[f"out_left_{i}" for i in kraus_out_modes]
        + [f"left_{i}" for i in kraus_in_modes],
    )
    kraus_conj = MMTensor(
        math.conj(kraus.tensor),
        axis_labels=[f"out_right_{i}" for i in kraus_out_modes]
        + [f"right_{i}" for i in kraus_in_modes],
    )

    # contract the kraus operator with the density matrix from the left and from the right.
    k_dm_k = kraus @ dm @ kraus_conj
    # now the leftover indices are in the order:
    # out_left_modes + uncontracted left indices + uncontracted right indices + out_right_modes

    # sort k_dm_k.axis_labels by the int at the end of each label, first left, then right
    N = k_dm_k.tensor.ndim // 2
    left = sorted(k_dm_k.axis_labels[:N], key=lambda x: int(x.split("_")[-1]))
    right = sorted(k_dm_k.axis_labels[N:], key=lambda x: int(x.split("_")[-1]))

    return k_dm_k.transpose(left + right).tensor


def apply_choi_to_dm(
    choi: ComplexTensor,
    dm: ComplexTensor,
    choi_in_modes: Sequence[int],
    choi_out_modes: Sequence[int] = None,
):
    r"""Applies a choi operator to a density matrix.
    It assumes that the density matrix is indexed as left_1, ..., left_n, right_1, ..., right_n.

    The choi operator has indices that contract with the density matrix (choi_in_modes) and indices that are left over (choi_out_modes).
    `choi` will contract choi_in_modes from the left and from the right with the density matrix.

    Args:
        choi (array): the choi operator to be applied
        dm (array): the density matrix to which the choi operator is applied
        choi_in_modes (list of ints): the input modes of the choi operator that contract with the density matrix
        choi_out_modes (list of ints): the output modes of the choi operator

    Returns:
        array: the resulting density matrix
    """
    if choi_out_modes is None:
        choi_out_modes = choi_in_modes
    if not set(choi_in_modes).issubset(range(dm.ndim // 2)):
        raise ValueError("choi_in_modes should be a subset of the density matrix indices.")

    # check that there are no repeated indices in kraus_in_modes and kraus_out_modes (separately)
    validate_contraction_indices(choi_in_modes, choi_out_modes, dm.ndim // 2, "choi")

    dm = MMTensor(
        dm,
        axis_labels=[f"in_left_{i}" for i in range(dm.ndim // 2)]
        + [f"in_right_{i}" for i in range(dm.ndim // 2)],
    )
    choi = MMTensor(
        choi,
        axis_labels=[f"out_left_{i}" for i in choi_out_modes]
        + [f"in_left_{i}" for i in choi_in_modes]
        + [f"out_right_{i}" for i in choi_out_modes]
        + [f"in_right_{i}" for i in choi_in_modes],
    )

    # contract the choi matrix with the density matrix.
    # now the leftover indices are in the order out_left_modes + out_right_modes + uncontracted left indices + uncontracted right indices
    choi_dm = choi @ dm

    # sort choi_dm.axis_labels by the int at the end of each label, first left, then right
    left_labels = [label for label in choi_dm.axis_labels if "left" in label]
    left = sorted(left_labels, key=lambda x: int(x.split("_")[-1]))
    right_labels = [label for label in choi_dm.axis_labels if "right" in label]
    right = sorted(right_labels, key=lambda x: int(x.split("_")[-1]))

    return choi_dm.transpose(left + right).tensor


def apply_choi_to_ket(choi, ket, choi_in_modes, choi_out_modes=None):
    r"""Applies a choi operator to a ket.
    It assumes that the ket is indexed as left_1, ..., left_n.

    The choi operator has indices that contract with the ket (choi_in_modes) and indices that are left over (choi_out_modes).
    `choi` will contract choi_in_modes from the left and from the right with the ket.

    Args:
        choi (array): the choi operator to be applied
        ket (array): the ket to which the choi operator is applied
        choi_in_modes (list of ints): the indices of the choi operator that contract with the ket
        choi_out_modes (list of ints): the indices of the choi operator that re leftover

    Returns:
        array: the resulting ket
    """
    if choi_out_modes is None:
        choi_out_modes = choi_in_modes

    if not set(choi_in_modes).issubset(range(ket.ndim)):
        raise ValueError("choi_in_modes should be a subset of the ket indices.")

    # check that there are no repeated indices in kraus_in_modes and kraus_out_modes (separately)
    validate_contraction_indices(choi_in_modes, choi_out_modes, ket.ndim, "choi")

    ket = MMTensor(ket, axis_labels=[f"left_{i}" for i in range(ket.ndim)])
    ket_dual = MMTensor(math.conj(ket.tensor), axis_labels=[f"right_{i}" for i in range(ket.ndim)])
    choi = MMTensor(
        choi,
        axis_labels=[f"out_left_{i}" for i in choi_out_modes]
        + [f"left_{i}" for i in choi_in_modes]
        + [f"out_right_{i}" for i in choi_out_modes]
        + [f"right_{i}" for i in choi_in_modes],
    )

    # contract the choi matrix with the ket and its dual, like choi @ |ket><ket|
    # now the leftover indices are in the order out_left_modes + out_right_modes + uncontracted left indices + uncontracted right indices
    choi_ket = choi @ ket @ ket_dual

    # sort choi_ket.axis_labels by the int at the end of each label, first left, then right
    left_labels = [label for label in choi_ket.axis_labels if "left" in label]
    left = sorted(left_labels, key=lambda x: int(x.split("_")[-1]))
    right_labels = [label for label in choi_ket.axis_labels if "right" in label]
    right = sorted(right_labels, key=lambda x: int(x.split("_")[-1]))

    return choi_ket.transpose(left + right).tensor


def contract_states(
    stateA, stateB, a_is_dm: bool, b_is_dm: bool, modes: List[int], normalize: bool
):
    r"""Contracts two states in the specified modes.
    Assumes that the modes of B are a subset of the modes of A.

    Args:
        stateA: the first state
        stateB: the second state
        a_is_dm: whether the first state is a density matrix.
        b_is_dm: whether the second state is a density matrix.
        modes: the modes on which to contract the states.
        normalize: whether to normalize the result

    Returns:
        Tensor: the contracted state tensor (subsystem of ``A``). Either ket or dm.
    """

    if a_is_dm:
        if b_is_dm:  # a DM, b DM
            dm = apply_choi_to_dm(choi=stateB, dm=stateA, choi_in_modes=modes, choi_out_modes=[])
        else:  # a DM, b ket
            dm = apply_kraus_to_dm(
                kraus=math.conj(stateB),
                dm=stateA,
                kraus_in_modes=modes,
                kraus_out_modes=[],
            )
    else:
        if b_is_dm:  # a ket, b DM
            dm = apply_kraus_to_dm(
                kraus=math.conj(stateA),
                dm=stateB,
                kraus_in_modes=modes,
                kraus_out_modes=[],
            )
        else:  # a ket, b ket
            ket = apply_kraus_to_ket(
                kraus=math.conj(stateB),
                ket=stateA,
                kraus_in_modes=modes,
                kraus_out_modes=[],
            )

    try:
        return dm / math.sum(math.all_diagonals(dm, real=False)) if normalize else dm
    except NameError:
        return ket / math.norm(ket) if normalize else ket


def normalize(fock: Tensor, is_dm: bool):
    r"""Returns the normalized ket state.

    Args:
        fock (Tensor): the state to be normalized
        is_dm (optioanl bool): whether the input tensor is a density matrix

    Returns:
        Tensor: the normalized state
    """
    if is_dm:
        return fock / math.sum(math.all_diagonals(fock, real=False))

    return fock / math.sum(math.norm(fock))


def norm(state: Tensor, is_dm: bool):
    r"""
    Returns the norm of a ket or the trace of the density matrix.
    Note that the "norm" is intended as the float number that is used to normalize the state,
    and depends on the representation. Hence different numbers for different representations
    of the same state (:math:`|amp|` for ``ket`` and :math:`|amp|^2` for ``dm``).
    """
    if is_dm:
        return math.sum(math.all_diagonals(state, real=True))

    return math.abs(math.norm(state))


def is_mixed_dm(dm):
    r"""Evaluates if a density matrix represents a mixed state."""
    cutoffs = dm.shape[: len(dm.shape) // 2]
    square = math.reshape(dm, (int(np.prod(cutoffs)), -1))
    return not np.isclose(math.sum(square * math.transpose(square)), 1.0)


def trace(dm, keep: List[int]):
    r"""Computes the partial trace of a density matrix.
    The indices of the density matrix are in the order (out0, ..., outN-1, in0, ..., inN-1).
    The indices to keep are a subset of the N 'out' indices
    (they count for the 'in' indices as well).

    Args:
        dm: the density matrix
        keep: the modes to keep (0-based)
    """
    dm = MMTensor(
        dm,
        axis_labels=[
            f"out_{i}" if i in keep else f"contract_{i}" for i in range(len(dm.shape) // 2)
        ]
        + [f"in_{i}" if i in keep else f"contract_{i}" for i in range(len(dm.shape) // 2)],
    )
    return dm.contract().tensor


@tensor_int_cache
def oscillator_eigenstate(q: Vector, cutoff: int) -> Tensor:
    r"""Harmonic oscillator eigenstate wavefunction `\psi_n(q) = <n|q>`.

    Args:
        q (Vector): a vector containing the q points at which the function is evaluated (units of \sqrt{\hbar})
        cutoff (int): maximum number of photons

    Returns:
        Tensor: a tensor of size ``len(q)*cutoff``. Each entry with index ``[i, j]`` represents the eigenstate evaluated
            with number of photons ``i`` evaluated at position ``q[j]``, i.e., `\psi_i(q_j)`.

    .. details::

        .. admonition:: Definition
            :class: defn

        The q-quadrature eigenstates are defined as

        .. math::

            \psi_n(x) = 1/sqrt[2^n n!](\frac{\omega}{\pi \hbar})^{1/4}
                \exp{-\frac{\omega}{2\hbar} x^2} H_n(\sqrt{\frac{\omega}{\pi}} x)

        where :math:`H_n(x)` is the (physicists) `n`-th Hermite polynomial.
    """
    omega_over_hbar = math.cast(1 / settings.HBAR, "float64")
    x_tensor = math.sqrt(omega_over_hbar) * math.cast(q, "float64")  # unit-less vector

    # prefactor term (\Omega/\hbar \pi)**(1/4) * 1 / sqrt(2**n)
    prefactor = (omega_over_hbar / np.pi) ** (1 / 4) * math.sqrt(
        math.pow(1 / 2, math.arange(0, cutoff))
    )

    # Renormalized physicist hermite polys: Hn / sqrt(n!)
    R = -np.array([[2 + 0j]])  # to get the physicist polys

    def f_hermite_polys(xi):
        poly = math.hermite_renormalized(
            R, 2 * math.astensor([xi], "complex128"), 1 + 0j, (cutoff,)
        )
        return math.cast(poly, "float64")

    hermite_polys = math.map_fn(f_hermite_polys, x_tensor)

    # (real) wavefunction
    psi = math.exp(-(x_tensor**2 / 2)) * math.transpose(prefactor * hermite_polys)
    return psi


@lru_cache
def estimate_dx(cutoff, period_resolution=20):
    r"""Estimates a suitable quadrature discretization interval `dx`. Uses the fact
    that Fock state `n` oscillates with angular frequency :math:`\sqrt{2(n + 1)}`,
    which follows from the relation

    .. math::

            \psi^{[n]}'(q) = q - sqrt(2*(n + 1))*\psi^{[n+1]}(q)

    by setting q = 0, and approximating the oscillation amplitude by `\psi^{[n+1]}(0)

    Ref: https://en.wikipedia.org/wiki/Hermite_polynomials#Hermite_functions

    Args
        cutoff (int): Fock cutoff
        period_resolution (int): Number of points used to sample one Fock
            wavefunction oscillation. Larger values yields better approximations
            and thus smaller `dx`.

    Returns
        (float): discretization value of quadrature
    """
    fock_cutoff_frequency = np.sqrt(2 * (cutoff + 1))
    fock_cutoff_period = 2 * np.pi / fock_cutoff_frequency
    dx_estimate = fock_cutoff_period / period_resolution
    return dx_estimate


@lru_cache
def estimate_xmax(cutoff, minimum=5):
    r"""Estimates a suitable quadrature axis length

    Args
        cutoff (int): Fock cutoff
        minimum (float): Minimum value of the returned xmax

    Returns
        (float): maximum quadrature value
    """
    if cutoff == 0:
        xmax_estimate = 3
    else:
        # maximum q for a classical particle with energy n=cutoff
        classical_endpoint = np.sqrt(2 * cutoff)
        # approximate probability of finding particle outside classical region
        excess_probability = 1 / (7.464 * cutoff ** (1 / 3))
        # Emperical factor that yields reasonable results
        A = 5
        xmax_estimate = classical_endpoint * (1 + A * excess_probability)
    return max(minimum, xmax_estimate)


@lru_cache
def estimate_quadrature_axis(cutoff, minimum=5, period_resolution=20):
    """Generates a suitable quadrature axis.

    Args
        cutoff (int): Fock cutoff
        minimum (float): Minimum value of the returned xmax
        period_resolution (int): Number of points used to sample one Fock
            wavefunction oscillation. Larger values yields better approximations
            and thus smaller dx.

    Returns
        (array): quadrature axis
    """
    xmax = estimate_xmax(cutoff, minimum=minimum)
    dx = estimate_dx(cutoff, period_resolution=period_resolution)
    xaxis = np.arange(-xmax, xmax, dx)
    xaxis = np.append(xaxis, xaxis[-1] + dx)
    xaxis = xaxis - np.mean(xaxis)  # center around 0
    return xaxis


def quadrature_distribution(
    state: Tensor,
    quadrature_angle: float = 0.0,
    x: Vector = None,
):
    r"""Given the ket or density matrix of a single-mode state, it generates the probability
    density distribution :math:`\tr [ \rho |x_\phi><x_\phi| ]`  where `\rho` is the
    density matrix of the state and |x_\phi> the quadrature eigenvector with angle `\phi`
    equal to ``quadrature_angle``.

    Args:
        state (Tensor): single mode state ket or density matrix
        quadrature_angle (float): angle of the quadrature basis vector
        x (Vector): points at which the quadrature distribution is evaluated

    Returns:
        tuple(Vector, Vector): coordinates at which the pdf is evaluated and the probability distribution
    """
    dims = len(state.shape)
    if dims > 2:
        raise ValueError(
            "Input state has dimension {state.shape}. Make sure is either a single-mode ket or dm."
        )

    is_dm = dims == 2
    cutoff = state.shape[0]

    if not np.isclose(quadrature_angle, 0.0):
        # rotate mode to the homodyne basis
        theta = -math.arange(cutoff) * quadrature_angle
        Ur = math.diag(math.make_complex(math.cos(theta), math.sin(theta)))
        state = (
            math.einsum("ij,jk,kl->il", Ur, state, math.dagger(Ur))
            if is_dm
            else math.matvec(Ur, state)
        )

    if x is None:
        x = np.sqrt(settings.HBAR) * math.new_constant(estimate_quadrature_axis(cutoff), "q_tensor")

    psi_x = math.cast(oscillator_eigenstate(x, cutoff), "complex128")
    pdf = (
        math.einsum("nm,nj,mj->j", state, psi_x, psi_x)
        if is_dm
        else math.abs(math.einsum("n,nj->j", state, psi_x)) ** 2
    )

    return x, math.cast(pdf, "float64")


def sample_homodyne(state: Tensor, quadrature_angle: float = 0.0) -> Tuple[float, float]:
    r"""Given a single-mode state, it generates the pdf of :math:`\tr [ \rho |x_\phi><x_\phi| ]`
    where `\rho` is the reduced density matrix of the state.

    Args:
        state (Tensor): ket or density matrix of the state being measured
        quadrature_angle (float): angle of the quadrature distribution

    Returns:
        tuple(float, float): outcome and probability of the outcome
    """
    dims = len(state.shape)
    if dims > 2:
        raise ValueError(
            "Input state has dimension {state.shape}. Make sure is either a single-mode ket or dm."
        )

    x, pdf = quadrature_distribution(state, quadrature_angle)
    probs = pdf * (x[1] - x[0])

    # draw a sample from the distribution
    pdf = math.Categorical(probs=probs, name="homodyne_dist")
    sample_idx = pdf.sample()
    homodyne_sample = math.gather(x, sample_idx)
    probability_sample = math.gather(probs, sample_idx)

    return homodyne_sample, probability_sample


@math.custom_gradient
def displacement(x, y, shape, tol=1e-15):
    r"""creates a single mode displacement matrix"""
    alpha = math.asnumpy(x) + 1j * math.asnumpy(y)

    if np.sqrt(x * x + y * y) > tol:
        gate = strategies.displacement(tuple(shape), alpha)
    else:
        gate = math.eye(max(shape), dtype="complex128")[: shape[0], : shape[1]]

    ret = math.astensor(gate, dtype=gate.dtype.name)
    if math.backend_name == "numpy":
        return ret

    def grad(dL_dDc):
        dD_da, dD_dac = strategies.jacobian_displacement(math.asnumpy(gate), alpha)
        dL_dac = np.sum(np.conj(dL_dDc) * dD_dac + dL_dDc * np.conj(dD_da))
        dLdx = 2 * np.real(dL_dac)
        dLdy = 2 * np.imag(dL_dac)
        return math.astensor(dLdx, dtype=x.dtype), math.astensor(dLdy, dtype=y.dtype)

    return ret, grad


@math.custom_gradient
def beamsplitter(theta: float, phi: float, shape: Sequence[int], method: str):
    r"""Creates a beamsplitter tensor with given cutoffs using a numba-based fock lattice strategy.

    Args:
        theta (float): transmittivity angle of the beamsplitter
        phi (float): phase angle of the beamsplitter
        cutoffs (int,int): cutoff dimensions of the two modes
    """
    if method == "vanilla":
        bs_unitary = strategies.beamsplitter(shape, math.asnumpy(theta), math.asnumpy(phi))
    elif method == "schwinger":
        bs_unitary = strategies.beamsplitter_schwinger(
            shape, math.asnumpy(theta), math.asnumpy(phi)
        )
    else:
        raise ValueError(
            f"Unknown beamsplitter method {method}. Options are 'vanilla' and 'schwinger'."
        )

    ret = math.astensor(bs_unitary, dtype=bs_unitary.dtype.name)
    if math.backend_name == "numpy":
        return ret

    def vjp(dLdGc):
        dtheta, dphi = strategies.beamsplitter_vjp(
            math.asnumpy(bs_unitary),
            math.asnumpy(math.conj(dLdGc)),
            math.asnumpy(theta),
            math.asnumpy(phi),
        )
        return math.astensor(dtheta, dtype=theta.dtype), math.astensor(dphi, dtype=phi.dtype)

    return ret, vjp


@math.custom_gradient
def squeezer(r, phi, shape):
    r"""creates a single mode squeezer matrix using a numba-based fock lattice strategy"""
    sq_unitary = strategies.squeezer(shape, math.asnumpy(r), math.asnumpy(phi))

    ret = math.astensor(sq_unitary, dtype=sq_unitary.dtype.name)
    if math.backend_name == "numpy":
        return ret

    def vjp(dLdGc):
        dr, dphi = strategies.squeezer_vjp(
            math.asnumpy(sq_unitary),
            math.asnumpy(math.conj(dLdGc)),
            math.asnumpy(r),
            math.asnumpy(phi),
        )
        return math.astensor(dr, dtype=r.dtype), math.astensor(dphi, phi.dtype)

    return ret, vjp


@math.custom_gradient
def squeezed(r, phi, shape):
    r"""creates a single mode squeezed state using a numba-based fock lattice strategy"""
    sq_ket = strategies.squeezed(shape, math.asnumpy(r), math.asnumpy(phi))

    ret = math.astensor(sq_ket, dtype=sq_ket.dtype.name)
    if math.backend_name == "numpy":
        return ret

    def vjp(dLdGc):
        dr, dphi = strategies.squeezed_vjp(
            math.asnumpy(sq_ket),
            math.asnumpy(math.conj(dLdGc)),
            math.asnumpy(r),
            math.asnumpy(phi),
        )
        return math.astensor(dr, dtype=r.dtype), math.astensor(dphi, phi.dtype)

    return ret, vjp



"""
This module contains helper functions that are used in
diagonal_amps.py, diagonal_grad.py, singleLeftoverMode_amps.py and singleLeftoverMode_grad.py
"""

import numpy as np
from numba import njit, int64
from numba.typed import Dict
import numba

SQRT = np.sqrt(np.arange(1000))  # saving the time to recompute square roots


@njit
def repeat_twice(params):
    """
    This function is equivalent to np.repeat(params,2), but runs faster.
    Args:
        params (1D array): [a,b,c,...]
    Returns:
        (1D array): [a,a,b,b,c,c,...]
    """
    pivot = np.zeros(2 * len(params), dtype=np.int64)
    for i, val in enumerate(params):
        pivot[2 * i] = val
        pivot[2 * i + 1] = val
    return pivot


@njit
def construct_dict_params(cutoffs, tuple_type, list_type):
    """
    Args:
        cutoffs (tuple): upper bounds for the number of photons in each mode
        tuple_type,list_type (numba types): numba types that need to be defined outside of numba compiled functions
    Returns:
        (typed Dict): all possible values for (a,b,c,...), grouped in lists according to their sum a+b+c+...
    """
    indices = Dict.empty(key_type=int64, value_type=list_type)
    for sum_params in range(sum(cutoffs)):
        indices[sum_params] = numba.typed.List.empty_list(tuple_type)

    for params in np.ndindex(cutoffs):
        indices[sum(params)].append(params)
    return indices



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Path functions"""

from numba import njit, typed, typeof, types
from numba.cpython.unsafe.tuple import tuple_setitem

BINOMIAL_PATHS_PYTHON = {}


@njit
def _binomial_subspace_basis(
    cutoffs: tuple[int, ...],
    weight: int,
    mode: int,
    basis_element: tuple[int, ...],
    basis: typed.List[tuple[int, ...]],
):
    r"""Step of the recursive function to generate all indices
    of a tensor with equal weight.
    If cutoffs is an empty tuple, the the basis element is appended to the list.
    Otherwise it loops over the values of the given mode, and it calls itself recursively
    to construct the rest of the basis elements.

    Arguments:
        cutoffs (tuple[int, ...]): the cutoffs of the tensor
        weight (int): the weight of the subspace
        mode (int): the mode to loop over
        basis_element (tuple[int, ...]): the current basis element to construct
        basis (list[tuple[int, ...]]): the list of basis elements to eventually append to
    """
    if mode == len(cutoffs):
        if weight == 0:  # ran out of photons to distribute
            basis.append(basis_element)
        return

    for photons in range(cutoffs[mode]):  # could be prange?
        if weight - photons >= 0:
            basis_element = tuple_setitem(basis_element, mode, photons)
            _binomial_subspace_basis(cutoffs, weight - photons, mode + 1, basis_element, basis)


@njit
def binomial_subspace_basis(cutoffs: tuple[int, ...], weight: int):
    r"""Returns all indices of a tensor with given weight.

    Arguments:
        cutoffs (tuple[int, ...]): the cutoffs of the tensor
        weight (int): the weight of the subspace

    Returns:
        list[tuple[int, ...]]: the list of basis elements of the subspace
    """
    basis = typed.List(
        [cutoffs]
    )  # this is just so that numba can infer the type, then we remove it
    _binomial_subspace_basis(cutoffs, weight, 0, cutoffs, basis)
    return basis[1:]  # remove the dummy element


def BINOMIAL_PATHS_NUMBA_n(modes):
    r"Creates a numba dictionary to store the paths and effectively cache them."
    return typed.Dict.empty(
        key_type=typeof(((0,) * modes, 0)),
        value_type=types.ListType(typeof((0,) * modes)),
    )



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Recurrencies for Fock-Bargmann amplitudes put together a strategy for
# enumerating the indices in a specific order and functions for calculating
# which neighbours to use in the calculation of the amplitude at a given index.
# In summary, they return the value of the amplitude at the given index by following
# a recipe made of two parts. The function to recompute A and b is determined by
# which neighbours are used.

"""Fock-Bargmann recurrence relation steps for Gaussian states."""

import numpy as np
from numba import njit, prange, types
from numba.cpython.unsafe.tuple import tuple_setitem

from mrmustard.math.lattice.neighbors import lower_neighbors
from mrmustard.math.lattice.pivots import first_available_pivot
from mrmustard.utils.typing import ComplexMatrix, ComplexTensor, ComplexVector

SQRT = np.sqrt(np.arange(100000))


@njit
def vanilla_step(
    G: ComplexTensor,
    A: ComplexMatrix,
    b: ComplexVector,
    index: tuple[int, ...],
) -> complex:
    r"""Fock-Bargmann recurrence relation step, vanilla version.
    This function returns the amplitude of the Gaussian tensor G
    at G[index]. It does not modify G.
    The necessary pivot and neighbours must have already been computed,
    as this step will read those values from G.

    Args:
        G (array or dict): fock amplitudes data store that supports getitem[tuple[int, ...]]
        A (array): A matrix of the Fock-Bargmann representation
        b (array): B vector of the Fock-Bargmann representation
        index (Sequence): index of the amplitude to calculate
    Returns:
        complex: the value of the amplitude at the given index
    """
    # get pivot
    i, pivot = first_available_pivot(index)

    # pivot contribution
    value_at_index = b[i] * G[pivot]

    # neighbors contribution
    for j, neighbor in lower_neighbors(pivot):
        value_at_index += A[i, j] * SQRT[pivot[j]] * G[neighbor]

    return value_at_index / SQRT[index[i]]


@njit
def vanilla_step_batch(
    G: ComplexTensor,
    A: ComplexMatrix,
    b: ComplexTensor,
    index: tuple[int, ...],
) -> complex:  # pragma: no cover
    r"""Fock-Bargmann recurrence relation step, vanilla batched version.
    This function returns the amplitude of the Gaussian tensor G
    at G[index]. It does not modify G.
    The necessary pivot and neighbours must have already been computed,
    as this step will read those values from G.
    Note that this function is different from vanilla_step with b is no longer a vector,
    it becomes a bathced vector with the batch dimension on the last index.

    Args:
        G (array or dict): fock amplitudes data store that supports getitem[tuple[int, ...]]
        A (array): A matrix of the Fock-Bargmann representation
        b (array): batched B vector of the Fock-Bargmann representation, the batch dimension is on the last index
        index (Sequence): index of the amplitude to calculate
    Returns:
        array: the value of the amplitude at the given index according to each batch on the last index
    """
    # get pivot
    i, pivot = first_available_pivot(index)

    # pivot contribution
    value_at_index = b[i] * G[pivot]

    # neighbors contribution
    for j, neighbor in lower_neighbors(pivot):
        value_at_index += A[i, j] * SQRT[pivot[j]] * G[neighbor]

    return value_at_index / SQRT[index[i]]


@njit
def vanilla_step_jacobian(
    G: ComplexTensor,
    A: ComplexMatrix,
    b: ComplexVector,
    index: tuple[int, ...],
    dGdA: ComplexTensor,
    dGdB: ComplexTensor,
) -> tuple[ComplexTensor, ComplexTensor]:
    r"""Jacobian contribution of a single Fock-Bargmann recurrence relation step, vanilla version.
    It updates the dGdB and dGdA tensors at the given index.

    Args:
        G (array or dict): fully computed store that supports __getitem__(index: tuple[int, ...])
        A (array): A matrix of the Fock-Bargmann representation
        b (array): B vector of the Fock-Bargmann representation
        c (complex): vacuum amplitude
        index (Sequence): index at which to compute the jacobian
        dGdB (array): gradient of G with respect to b (partially computed)
        dGdA (array): gradient of G with respect to A (partially computed)
    Returns:
        tuple[array, array]: the dGdB and dGdA tensors updated at the given index
    """
    # index -> pivot
    i, pivot = first_available_pivot(index)

    # pivot contribution
    dGdB[index] += b[i] * dGdB[pivot] / SQRT[index[i]]
    dGdB[index + (i,)] += G[pivot] / SQRT[index[i]]
    dGdA[index] += b[i] * dGdA[pivot] / SQRT[index[i]]

    # neighbors contribution
    for j, neighbor in lower_neighbors(pivot):
        dGdB[index] += A[i, j] * dGdB[neighbor] * SQRT[pivot[j]] / SQRT[index[i]]
        dGdA[index] += A[i, j] * dGdA[neighbor] * SQRT[pivot[j]] / SQRT[index[i]]
        dGdA[index + (i, j)] += G[neighbor] * SQRT[pivot[j]] / SQRT[index[i]]

    return dGdA, dGdB


@njit
def vanilla_step_grad(
    G: ComplexTensor,
    index: tuple[int, ...],
    dA: ComplexMatrix,
    db: ComplexVector,
) -> tuple[ComplexMatrix, ComplexVector]:
    r"""Gradient with respect to A and b of a single Fock-Bargmann recurrence relation step,
    vanilla version. dA and db can be used to update the dGdB and dGdA tensors at `index`,
    or as part of the contraction with the gradient of the Gaussian tensor G.
    Note that the gradient depends only on G and not on the values of A and b.

    Args:
        G (array or dict): fully computed store that supports __getitem__(index: tuple[int, ...])
        index (Sequence): index of the amplitude to calculate the gradient of
        dA (array): empty array to store the gradient of G[index] with respect to A
        db (array): empty array to store the gradient of G[index] with respect to B
    Returns:
        tuple[array, array]: the updated dGdB and dGdA tensors
    """
    for i in range(len(db)):  # pylint: disable=consider-using-enumerate
        pivot_i = tuple_setitem(index, i, index[i] - 1)
        db[i] = SQRT[index[i]] * G[pivot_i]
        dA[i, i] = 0.5 * SQRT[index[i] * pivot_i[i]] * G[tuple_setitem(pivot_i, i, pivot_i[i] - 1)]
        for j in range(i + 1, len(db)):
            dA[i, j] = SQRT[index[i] * pivot_i[j]] * G[tuple_setitem(pivot_i, j, pivot_i[j] - 1)]

    return dA, db


@njit
def vanilla_step_dict(
    data: types.DictType, A: ComplexMatrix, b: ComplexVector, index: tuple[int, ...]
) -> complex:
    r"""Fock-Bargmann recurrence relation step, vanilla version with numba dict.
    This function calculates the index `index` of the Gaussian tensor `G`.
    The appropriate pivot and neighbours must exist.

    Args:
        data: dict(tuple[int,...], complex): fock amplitudes numba dict
        A (array): A matrix of the Fock-Bargmann representation
        b (array): b vector of the Fock-Bargmann representation
        index (Sequence): index of the amplitude to calculate
    Returns:
        complex: the value of the amplitude at the given index
    """
    # index -> pivot
    i, pivot = first_available_pivot(index)

    # calculate value at index: pivot contribution
    denom = SQRT[pivot[i] + 1]
    value_at_index = b[i] / denom * data[pivot]

    # neighbors contribution
    for j, neighbor in lower_neighbors(pivot):
        value_at_index += A[i, j] / denom * SQRT[pivot[j]] * data.get(neighbor, 0.0 + 0.0j)

    return value_at_index


@njit
def binomial_step(
    G: ComplexTensor, A: ComplexMatrix, b: ComplexVector, subspace_indices: list[tuple[int, ...]]
) -> tuple[ComplexTensor, float]:
    r"""Computes a whole subspace of the ``G`` tensor at the indices in
    ``subspace_indices`` (a subspace is such that `sum(index) = const`).
    It updates the tensor ``G``.
    It returns the updated tensor and the probability of the subspace.

    Args:
        G (np.ndarray): Tensor filled up to the previous subspace
        A (np.ndarray): A matrix of the Fock-Bargmann representation
        b (np.ndarray): B vector of the Fock-Bargmann representation
        subspace_indices (list[tuple[int, ...]]): list of indices to be updated

    Returns:
        tuple[np.ndarray, float]: updated tensor and probability of the subspace
    """
    norm = 0.0

    for i in prange(len(subspace_indices)):
        value = vanilla_step(G, A, b, subspace_indices[i])
        G[subspace_indices[i]] = value
        norm = norm + np.abs(value) ** 2

    return G, norm


@njit
def binomial_step_dict(
    G: types.DictType, A: ComplexMatrix, b: ComplexVector, subspace_indices: list[tuple[int, ...]]
) -> tuple[types.DictType, float]:
    r"""Computes a whole subspace of the ``G`` dict at the indices in
    ``subspace_indices`` (a subspace is such that `sum(index) = const`).
    It updates the dict ``G``. It returns the updated G dict
    and the total probability of the subspace.

    Args:
        G (types.DictType): Dictionary filled up to the previous subspace
        A (np.ndarray): A matrix of the Fock-Bargmann representation
        b (np.ndarray): B vector of the Fock-Bargmann representation
        subspace_indices (list[tuple[int, ...]]): list of indices to be updated

    Returns:
        tuple[types.DictType, float]: updated dictionary and probability of the subspace
    """
    prob = 0.0

    for i in prange(len(subspace_indices)):
        value = vanilla_step_dict(G, A, b, subspace_indices[i])
        G[subspace_indices[i]] = value
        prob = prob + np.abs(value) ** 2

    return G, prob



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module containing all base type annotations."""
__all__ = [
    "RealVector",
    "ComplexVector",
    "IntVector",
    "UIntVector",
    "RealMatrix",
    "ComplexMatrix",
    "IntMatrix",
    "UIntMatrix",
    "RealTensor",
    "ComplexTensor",
    "IntTensor",
    "UIntTensor",
    "Batch",
    "Scalar",
    "Vector",
    "Matrix",
    "Tensor",
    "Trainable",
]
from typing import (
    Iterator,
    Protocol,
    Tuple,
    TypeVar,
    Union,
    runtime_checkable,
)

import numpy as np

R = TypeVar("R", np.float16, np.float32, np.float64)
C = TypeVar("C", np.complex64, np.complex128)
Z = TypeVar("Z", np.int16, np.int32, np.int64)
N = TypeVar("N", np.uint16, np.uint32, np.uint64)

Scalar = Union[R, C, Z, N]
Vector = np.ndarray[Tuple[int], Scalar]
Matrix = np.ndarray[Tuple[int, int], Scalar]
Tensor = np.ndarray[Tuple[int, ...], Scalar]

RealVector = np.ndarray[Tuple[int], R]
ComplexVector = np.ndarray[Tuple[int], C]
IntVector = np.ndarray[Tuple[int], Z]
UIntVector = np.ndarray[Tuple[int], N]

RealMatrix = np.ndarray[Tuple[int, int], R]
ComplexMatrix = np.ndarray[Tuple[int, int], C]
IntMatrix = np.ndarray[Tuple[int, int], Z]
UIntMatrix = np.ndarray[Tuple[int, int], N]

RealTensor = np.ndarray[Tuple[int, ...], R]
ComplexTensor = np.ndarray[Tuple[int, ...], C]
IntTensor = np.ndarray[Tuple[int, ...], Z]
UIntTensor = np.ndarray[Tuple[int, ...], N]


# Revisit when requiring python 3.12 (see PEP 695)
T_co = TypeVar(
    "T_co",
    RealVector,
    ComplexVector,
    IntVector,
    UIntVector,
    RealMatrix,
    ComplexMatrix,
    IntMatrix,
    UIntMatrix,
    RealTensor,
    ComplexTensor,
    IntTensor,
    UIntTensor,
    covariant=True,
)

Trainable = TypeVar("Trainable")


@runtime_checkable
class Batch(Protocol[T_co]):
    r"""Anything that can iterate over objects of type T_co."""

    def __iter__(self) -> Iterator[T_co]: ...



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the classes to describe sets of parameters."""

from typing import Sequence, Union

from mrmustard.math.backend_manager import BackendManager

from .parameters import Constant, Variable

math = BackendManager()

__all__ = [
    "ParameterSet",
]


class ParameterSet:
    r"""
    A set of parameters.

    ``ParameterSet`` can store both constant and variable parameters. It provides fast access to
    both classes of parameters, as well as to their names.

    .. code::

      const1 = Constant(1.2345, "const1")
      const2 = Constant(2.3456, "const2")
      var1 = Variable(3.4567, "var1")

      ps = ParameterSet()
      ps.add_parameter(const1)
      ps.add_parameter(const2)
      ps.add_parameter(var1)

      ps.names  # returns `["const1", "const2", "var1"]`
      ps.constants  # returns `{"const1": const1, "const2": const2}`
      ps.variable  # returns `{"var1": var1}`
    """

    def __init__(self):
        self._names: list[str] = []
        self._constants: dict[str, Constant] = {}
        self._variables: dict[str, Variable] = {}

    @property
    def constants(self) -> dict[str, Constant]:
        r"""
        The constant parameters in this parameter set.
        """
        return self._constants

    @property
    def variables(self) -> dict[str, Variable]:
        r"""
        The variable parameters in this parameter set.
        """
        return self._variables

    @property
    def names(self) -> Sequence[str]:
        r"""
        The names of all the parameters in this parameter set, in the order in which they
        were added.
        """
        return self._names

    def add_parameter(self, parameter: Union[Constant, Variable]) -> None:
        r"""
        Adds a parameter to this parameter set.

        Args:
            parameter: A constant or variable parameter.

        Raises:
            ValueError: If this parameter set already contains a parameter with the same
                name as that of the given parameter.
        """
        name = parameter.name

        if name in self.names:
            msg = f"A parameter with name ``{name}`` is already part of this parameter set."
            raise ValueError(msg)
        self._names.append(name)

        # updates dictionary and dynamically creates an attribute
        if isinstance(parameter, Constant):
            self.constants[name] = parameter
            self.__dict__[name] = self.constants[name]
        elif isinstance(parameter, Variable):
            self.variables[parameter.name] = parameter
            self.__dict__[name] = self.variables[name]

    def tagged_variables(self, tag: str) -> dict[str, Variable]:
        r"""
        Returns a dictionary whose keys are tagged names of the variables in this parameter set, and whose
        values are the variables in this parameter set. Tagging is done by prepending the string ``f"{tag}"/``
        to variables' original names.
        """
        ret = {}
        for k, v in self.variables.items():
            ret[f"{tag}/{k}"] = v
        return ret

    def to_string(self, decimals: int) -> str:
        r"""
        Returns a string representation of the parameter values, separated by commas and rounded
        to the specified number of decimals.

        Args:
            decimals (int): number of decimals to round to

        Returns:
            str: string representation of the parameter values
        """
        strings = []
        for name in self.names:
            param = self.constants.get(name) or self.variables.get(name)
            if len(param.value.shape) == 0:  # don't show arrays
                string = str(math.asnumpy(math.round(param.value, decimals)))
            else:
                string = f"{name}"
            strings.append(string)
        return ", ".join(strings)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Functions and classes for tensor networks."""

from __future__ import annotations

from typing import Optional
from opt_einsum import contract as opt_contract

import networkx as nx
import numpy as np
import matplotlib.pyplot as plt

from .tensors import Wire, Tensor


def connect(wire1: Wire, wire2: Wire, dim: Optional[int] = None):
    r"""Connects two wires in a tensor network.

    Args:
        wire1: The first wire.
        wire2: The second wire.
        dim: The dimension of the contraction.

    Raises:
        ValueError: If one or both of the wires are already connected.
    """
    if wire1.is_connected or wire2.is_connected:
        msg = "Tried to connect wires that are already connected."
        raise ValueError(msg)

    if dim:
        wire1.dim = dim
        wire2.dim = dim

    wire1.is_connected = True
    wire2.is_connected = True

    wire1.contraction_id = wire2.contraction_id


def contract(tensors: list[Tensor], default_dim: int):
    r"""Contract a list of tensors.

    Args:
        tensors: The tensors to contract.
        dim: The default dimension of the contractions.

    Returns:
        The contracted tensor.
    """
    opt_einsum_args = []
    for t in tensors:
        shape = t.shape(default_dim=default_dim, out_in=True)
        opt_einsum_args.append(t.value(shape=shape))
        opt_einsum_args.append([w.contraction_id for w in t.wires])
    return opt_contract(*opt_einsum_args)


def draw(tensors: list[Tensor], layout: str = "spring_layout", figsize: tuple[int, int] = (10, 6)):
    r"""Draws a tensor network.

    Args:
        tensors: The tensors to draw.
        layout: The layout method. Must be one of the methods in ``nx.drawing.layout``.
        figsize: The size of the returned figure.

    Returns:
        A figure showing the tensor network.
    """
    try:
        fn_layout = getattr(nx.drawing.layout, layout)
    except AttributeError:
        msg = f"Invalid layout {layout}."
        # pylint: disable=raise-missing-from
        raise ValueError(msg)

    # initialize empty lists and dictionaries used to store metadata
    tensor_labels = {}
    mode_labels = {}
    node_size = []
    node_color = []

    # initialize three graphs--one to store nodes and edges, two to keep track of arrows
    graph = nx.Graph()
    arrows_in = nx.Graph()
    arrows_out = nx.Graph()

    for idx, tensor in enumerate(tensors):
        tensor_id = tensor.name + str(idx)
        graph.add_node(tensor_id)
        tensor_labels[tensor_id] = tensor.name
        mode_labels[tensor_id] = ""
        node_size.append(150)
        node_color.append("red")

        for wire in tensor.wires:
            wire_id = wire.contraction_id
            if wire_id not in graph.nodes:
                node_size.append(0)
                node_color.append("white")
                tensor_labels[wire_id] = ""
                mode_labels[wire_id] = wire.mode

            graph.add_node(wire_id)
            graph.add_edge(wire_id, tensor_id)
            if wire.is_input:
                arrows_in.add_edge(tensor_id, wire_id)
            else:
                arrows_out.add_edge(tensor_id, wire_id)

    pos = fn_layout(graph)
    pos_labels = {k: v + np.array([0.0, 0.05]) for k, v in pos.items()}

    fig = plt.figure(figsize=figsize)
    nx.draw_networkx_nodes(
        graph, pos, edgecolors="gray", alpha=0.9, node_size=node_size, node_color=node_color
    )
    nx.draw_networkx_edges(graph, pos, edge_color="lightgreen", width=4, alpha=0.6)
    nx.draw_networkx_edges(
        arrows_in,
        pos,
        edge_color="darkgreen",
        width=0.5,
        arrows=True,
        arrowsize=10,
        arrowstyle="<|-",
    )
    nx.draw_networkx_edges(
        arrows_out,
        pos,
        edge_color="darkgreen",
        width=0.5,
        arrows=True,
        arrowsize=10,
        arrowstyle="-|>",
    )
    nx.draw_networkx_labels(
        graph,
        pos=pos_labels,
        labels=tensor_labels,
        font_size=12,
        font_color="black",
        font_family="serif",
    )
    nx.draw_networkx_labels(
        graph,
        pos=pos_labels,
        labels=mode_labels,
        font_size=12,
        font_color="black",
        font_family="FreeMono",
    )

    plt.title("Mr Mustard Network")
    plt.show()
    return fig



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the classes to describe constant and variable parameters used in Mr Mustard."""

from typing import Callable, Optional, Tuple

from mrmustard.math.backend_manager import BackendManager

math = BackendManager()


__all__ = ["Constant", "Variable"]


# ~~~~~~~~~
# Functions
# ~~~~~~~~~


def update_symplectic(grads_and_vars, symplectic_lr: float):
    r"""
    Updates the symplectic parameters using the given symplectic gradients.

    Implemented from:
        Wang J, Sun H, Fiori S. A Riemannian-steepest-descent approach
        for optimization on the real symplectic group.
        Mathematical Methods in the Applied Sciences. 2018 Jul 30;41(11):4273-86.
    """
    for dS_euclidean, S in grads_and_vars:
        Y = math.euclidean_to_symplectic(S, dS_euclidean)
        YT = math.transpose(Y)
        new_value = math.matmul(
            S, math.expm(-symplectic_lr * YT) @ math.expm(-symplectic_lr * (Y - YT))
        )
        math.assign(S, new_value)


def update_orthogonal(grads_and_vars, orthogonal_lr: float):
    r"""Updates the orthogonal parameters using the given orthogonal gradients.

    Implemented from:
        Y Yao, F Miatto, N Quesada - arXiv preprint arXiv:2209.06069, 2022.
    """
    for dO_euclidean, O in grads_and_vars:
        Y = math.euclidean_to_unitary(O, math.real(dO_euclidean))
        new_value = math.matmul(O, math.expm(-orthogonal_lr * Y))
        math.assign(O, new_value)


def update_unitary(grads_and_vars, unitary_lr: float):
    r"""Updates the unitary parameters using the given unitary gradients.

    Implemented from:
        Y Yao, F Miatto, N Quesada - arXiv preprint arXiv:2209.06069, 2022.
    """
    for dU_euclidean, U in grads_and_vars:
        Y = math.euclidean_to_unitary(U, dU_euclidean)
        new_value = math.matmul(U, math.expm(-unitary_lr * Y))
        math.assign(U, new_value)


def update_euclidean(grads_and_vars, euclidean_lr: float):
    """Updates the parameters using the euclidian gradients."""
    math.euclidean_opt.lr = euclidean_lr
    math.euclidean_opt.apply_gradients(grads_and_vars)


# ~~~~~~~
# Classes
# ~~~~~~~


class Constant:
    r"""
    A parameter with a constant, immutable value.

    .. code::

      my_const = Constant(1, "my_const")

    Args:
        value: The value of this constant.
        name: The name of this constant.
    """

    def __init__(self, value: any, name: str):
        if math.from_backend(value) and not math.is_trainable(value):
            self._value = value
        elif type(value) in [list, int, float]:
            self._value = math.new_constant(value, name)
        else:
            self._value = math.new_constant(value, name, value.dtype)

        self._name = name

    @property
    def name(self) -> str:
        r"""
        The name of this constant.
        """
        return self._name

    @property
    def value(self) -> any:
        r"""
        The value of this constant.
        """
        return self._value

    def __mul__(self, value):
        return type(self)(value=value * self.value, name=self.name)

    def __rmul__(self, value):
        return type(self)(value=self.value * value, name=self.name)


# pylint: disable=too-many-instance-attributes
class Variable:
    r"""
    A parameter whose value can change.

    .. code::

      my_var = Variable(1, "my_var")

    Args:
        value: The value of this variable.
        name: The name of this variable.
        bounds: The numerical bounds of this variable.
        update_fn: The function used to update this variable during training.
    """

    def __init__(
        self,
        value: any,
        name: str,
        bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        update_fn: Callable = update_euclidean,
    ):
        self._value = self._get_value(value, bounds, name)
        self._name = name
        self._bounds = bounds
        self._update_fn = update_fn

    def _get_value(self, value, bounds, name):
        r"""
        Returns a variable from given ``value``, ``bounds``, and ``name``.
        """
        if math.from_backend(value) and math.is_trainable(value):
            return value
        elif type(value) in [list, int, float]:
            return math.new_variable(value, bounds, name)
        else:
            return math.new_variable(value, bounds, name, value.dtype)

    @property
    def bounds(self) -> Tuple[Optional[float], Optional[float]]:
        r"""
        The numerical bounds of this variable.
        """
        return self._bounds

    @property
    def name(self) -> str:
        r"""
        The name of this constant.
        """
        return self._name

    @property
    def update_fn(self) -> Optional[Callable]:
        r"""
        The function used to update this variable during training.
        """
        return self._update_fn

    @update_fn.setter
    def update_fn(self, value):
        self._update_fn = value

    @property
    def value(self) -> any:
        r"""
        The value of this variable.
        """
        return self._value

    @value.setter
    def value(self, value):
        self._value = self._get_value(value, self.bounds, self.name)

    @staticmethod
    def orthogonal(
        value: Optional[any],
        name: str,
        bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        N: int = 1,
    ):
        r"""
        Initializes a variable with ``update_fn`` for orthogonal optimization.

        Args:
            value: The value of the returned variable. If ``None``, a random orthogonal
                matrix of dimension ``N`` is initialized.
            name: The name of the returned variable.
            bounds: The numerical bounds of the returned variable.
            N: The dimension of the random orthogonal matrix. This value is ignored if
                ``value`` is not ``None``.

        Returns:
            A variable with ``update_fn`` for orthogonal optimization.
        """
        value = value or math.random_orthogonal(N)
        return Variable(value, name, bounds, update_orthogonal)

    @staticmethod
    def symplectic(
        value: any,
        name: str,
        bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        N: int = 1,
    ):
        r"""
        Initializes a variable with ``update_fn`` for simplectic optimization.

        Args:
            value: The value of the returned variable. If ``None``, a random symplectic
                matrix of dimension ``N`` is initialized.
            name: The name of the returned variable.
            bounds: The numerical bounds of the returned variable.
            N: The dimension of the random symplectic matrix. This value is ignored if
                ``value`` is not ``None``.

        Returns:
            A variable with ``update_fn`` for simplectic optimization.
        """
        value = value or math.random_symplectic(N)
        return Variable(value, name, bounds, update_symplectic)

    @staticmethod
    def unitary(
        value: any,
        name: str,
        bounds: Tuple[Optional[float], Optional[float]] = (None, None),
        N: int = 1,
    ):
        r"""
        Initializes a variable with ``update_fn`` for unitary optimization.

        Args:
            value: The value of the returned variable. If ``None``, a random unitary
                matrix of dimension ``N`` is initialized.
            name: The name of the returned variable.
            bounds: The numerical bounds of the returned variable.
            N: The dimension of the random unitary matrix. This value is ignored if
                ``value`` is not ``None``.

        Returns:
            A variable with ``update_fn`` for unitary optimization.
        """
        value = value or math.random_unitary(N)
        return Variable(value, name, bounds, update_unitary)

    def __mul__(self, value):
        return type(self)(
            value=value * self.value,
            name=self.name,
            bounds=self.bounds,
            update_fn=self.update_fn,
        )

    def __rmul__(self, value):
        return type(self)(
            value=self.value * value,
            name=self.name,
            bounds=self.bounds,
            update_fn=self.update_fn,
        )



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from typing import Any, Union, Optional

import numpy as np

from mrmustard import math
from mrmustard.utils.argsort import argsort_gen
from mrmustard.utils.typing import (
    Batch,
    ComplexMatrix,
    ComplexTensor,
    ComplexVector,
    Matrix,
    Scalar,
    Tensor,
    Vector,
)

__all__ = ["Ansatz", "PolyExpBase", "PolyExpAnsatz"]


class Ansatz(ABC):
    r"""An Ansatz is a function over a continuous and/or discrete domain.
    It supports many mathematical operations such as addition, subtraction,
    multiplication, division, negation, equality, etc.

    Note that n-dimensional arrays are like functions defined over an integer lattice of points,
    so this class also works for e.g. the Fock representation.

    This class is abstract. Concrete Ansatz classes will have to implement the
    ``__call__``, ``__mul__``, ``__add__``, ``__sub__``, ``__neg__`` and ``__eq__`` methods.
    """

    @abstractmethod
    def __neg__(self) -> Ansatz:
        r"""
        Negates this ansatz.
        """
        ...

    @abstractmethod
    def __eq__(self, other: Ansatz) -> bool:
        r"""
        Whether this ansatz is equal to another ansatz.
        """
        ...

    @abstractmethod
    def __add__(self, other: Ansatz) -> Ansatz:
        r"""
        Sums this ansatz to another ansatz.
        """
        ...

    def __sub__(self, other: Ansatz) -> Ansatz:
        r"""
        Subtracts other from this ansatz.
        """
        try:
            return self.__add__(-other)
        except AttributeError as e:
            raise TypeError(f"Cannot subtract {self.__class__} and {other.__class__}.") from e

    @abstractmethod
    def __call__(self, point: Any) -> Scalar:
        r"""
        Evaluates this ansatz at a given point in the domain.
        """
        ...

    @abstractmethod
    def __truediv__(self, other: Union[Scalar, Ansatz]) -> Ansatz:
        r"""
        Divides this ansatz by another ansatz or by a scalar.
        """
        ...

    @abstractmethod
    def __mul__(self, other: Union[Scalar, Ansatz]) -> Ansatz:
        r"""
        Multiplies this ansatz by another ansatz.
        """
        ...

    def __rmul__(self, other: Scalar) -> Ansatz:
        r"""
        Multiplies this ansatz by a scalar.
        """
        return self.__mul__(other=other)


class PolyExpBase(Ansatz):
    r"""
    A family of Ansatze parametrized by a triple of a matrix, a vector and an array.
    For example, the Bargmann representation :math:`c\:\textrm{exp}(z A z / 2 + b z)` is of this
    form (where ``A``, ``b``, ``c`` is the triple), or the Wigner representation
    (where ``Sigma``, ``mu``, ``1`` is the triple).

    Note that this class is not initializable (despite having an initializer) because it does
    not implement all the abstract methods of ``Ansatz``, and it is in fact more general.
    Concrete ansatze that inherit from this class need to implement ``__call__``,
    ``__mul__`` and ``__matmul__``, which are representation-specific.

    Note that the arguments are expected to be batched, i.e. to have a batch dimension
    or to be an iterable. This is because this class also provides the linear superposition
    functionality by implementing the ``__add__`` method, which concatenates the batch dimensions.

    As this can blow up the number of terms in the representation, it is recommended to
    run the `simplify()` method after adding terms together, which will combine together
    terms that have the same exponential part.

    Args:
        mat: the matrix-like data
        vec: the vector-like data
        array: the array-like data
    """

    def __init__(self, mat: Batch[Matrix], vec: Batch[Vector], array: Batch[Tensor]):
        self.mat = math.atleast_3d(math.astensor(mat))
        self.vec = math.atleast_2d(math.astensor(vec))
        self.array = math.atleast_1d(math.astensor(array))
        self.batch_size = self.mat.shape[0]
        self.dim = self.mat.shape[-1]
        self._simplified = False

    def __neg__(self) -> PolyExpBase:
        return self.__class__(self.mat, self.vec, -self.array)

    def __eq__(self, other: PolyExpBase) -> bool:
        return self._equal_no_array(other) and np.allclose(self.array, other.array, atol=1e-10)

    def _equal_no_array(self, other: PolyExpBase) -> bool:
        self.simplify()
        other.simplify()
        return np.allclose(self.vec, other.vec, atol=1e-10) and np.allclose(
            self.mat, other.mat, atol=1e-10
        )

    def __add__(self, other: PolyExpBase) -> PolyExpBase:
        combined_matrices = math.concat([self.mat, other.mat], axis=0)
        combined_vectors = math.concat([self.vec, other.vec], axis=0)
        combined_arrays = math.concat([self.array, other.array], axis=0)
        # note output is not simplified
        return self.__class__(combined_matrices, combined_vectors, combined_arrays)

    @property
    def degree(self) -> int:
        if self.array.ndim == 1:
            return 0
        return self.array.shape[-1] - 1

    def simplify(self) -> None:
        r"""
        Simplifies the representation by combining together terms that have the same
        exponential part, i.e. two terms along the batch are considered equal if their
        matrix and vector are equal. In this case only one is kept and the arrays are added.

        Does not run if the representation has already been simplified, so it is safe to call.
        """
        if self._simplified:
            return
        indices_to_check = set(range(self.batch_size))
        removed = []
        while indices_to_check:
            i = indices_to_check.pop()
            for j in indices_to_check.copy():
                if np.allclose(self.mat[i], self.mat[j]) and np.allclose(self.vec[i], self.vec[j]):
                    self.array = math.update_add_tensor(self.array, [[i]], [self.array[j]])
                    indices_to_check.remove(j)
                    removed.append(j)
        to_keep = [i for i in range(self.batch_size) if i not in removed]
        self.mat = math.gather(self.mat, to_keep, axis=0)
        self.vec = math.gather(self.vec, to_keep, axis=0)
        self.array = math.gather(self.array, to_keep, axis=0)
        self._simplified = True

    def simplify_v2(self) -> None:
        r"""
        A different implementation that orders the batch dimension first.
        """
        if self._simplified:
            return
        self._order_batch()
        to_keep = [d0 := 0]
        mat, vec = self.mat[d0], self.vec[d0]
        for d in range(1, self.batch_size):
            if np.allclose(mat, self.mat[d]) and np.allclose(vec, self.vec[d]):
                self.array = math.update_add_tensor(self.array, [[d0]], [self.array[d]])
            else:
                to_keep.append(d)
                d0 = d
                mat, vec = self.mat[d0], self.vec[d0]
        self.mat = math.gather(self.mat, to_keep, axis=0)
        self.vec = math.gather(self.vec, to_keep, axis=0)
        self.array = math.gather(self.array, to_keep, axis=0)
        self._simplified = True

    def _order_batch(self):
        r"""This method orders the batch dimension by the lexicographical order of the
        flattened arrays (mat, vec, array). This is a very cheap way to enforce
        an ordering of the batch dimension, which is useful for simplification and for
        determining (in)equality between two Bargmann representations."""
        generators = [
            itertools.chain(
                math.asnumpy(self.vec[i]).flat,
                math.asnumpy(self.mat[i]).flat,
                math.asnumpy(self.array[i]).flat,
            )
            for i in range(self.batch_size)
        ]
        sorted_indices = argsort_gen(generators)
        self.mat = math.gather(self.mat, sorted_indices, axis=0)
        self.vec = math.gather(self.vec, sorted_indices, axis=0)
        self.array = math.gather(self.array, sorted_indices, axis=0)


class PolyExpAnsatz(PolyExpBase):
    r"""
    Represents the ansatz function:

        :math:`F(z) = \sum_i \textrm{poly}_i(z) \textrm{exp}(z^T A_i z / 2 + z^T b_i)`

    where each :math:`poly_i` is a polynomial in ``z`` that can be expressed as

        :math:`\textrm{poly}_i(z) = \sum_k c^(i)_k z^k`,

    with ``k`` being a multi-index. The matrices :math:`A_i` and vectors :math:`b_i` are
    parameters of the exponential terms in the ansatz, and :math:`z` is a vector of variables.

    Args:
        A: The list of square matrices :math:`A_i`
        b: The list of vectors :math:`b_i`
        c: The array of coefficients for the polynomial terms in the ansatz.

    .. code-block::

        A = [np.array([[1.0, 0.0], [0.0, 1.0]])]
        b = [np.array([1.0, 1.0])]
        c = [np.array(1.0)]
        F = PolyExpAnsatz(A, b, c)
        z = np.array([1.0, 2.0])
        print(F(z))  # prints the value of F at z

    """

    def __init__(
        self,
        A: Optional[Batch[Matrix]] = None,
        b: Optional[Batch[Vector]] = None,
        c: Batch[Tensor | Scalar] = [1.0],
        name: str = "",
    ):
        self.name = name
        if A is None and b is None:
            raise ValueError("Please provide either A or b.")
        dim = b[0].shape[-1] if A is None else A[0].shape[-1]
        A = A if A is not None else np.zeros((len(b), dim, dim), dtype=b[0].dtype)
        b = b if b is not None else np.zeros((len(A), dim), dtype=A[0].dtype)
        super().__init__(mat=A, vec=b, array=c)

    @property
    def A(self) -> Batch[ComplexMatrix]:
        r"""
        The list of square matrices :math:`A_i`.
        """
        return self.mat

    @property
    def b(self) -> Batch[ComplexVector]:
        r"""
        The list of vectors :math:`b_i`.
        """
        return self.vec

    @property
    def c(self) -> Batch[ComplexTensor]:
        r"""
        The array of coefficients for the polynomial terms in the ansatz.
        """
        return self.array

    def __call__(self, z: Batch[Vector]) -> Scalar:
        r"""
        Value of this ansatz at ``z``.

        Args:
            z: point in C^n where the function is evaluated

        Returns:
            The value of the function.
        """
        z = np.atleast_2d(z)  # shape (..., n)
        zz = np.einsum("...a,...b->...ab", z, z)[..., None, :, :]  # shape (..., 1, n, n))
        A_part = 0.5 * math.sum(
            zz * self.A, axes=[-1, -2]
        )  # sum((...,1,n,n) * (b,n,n), [-1,-2]) ~ (...,b)
        b_part = np.sum(z[..., None, :] * self.b, axis=-1)  # sum((...,1,n) * (b,n), -1) ~ (...,b)
        exp_sum = np.exp(A_part + b_part)  # (..., b)
        result = exp_sum * self.c  # (..., b)
        val = np.sum(result, axis=-1)  # (...)
        return val

    def __mul__(self, other: Union[Scalar, PolyExpAnsatz]) -> PolyExpAnsatz:
        r"""Multiplies this ansatz by a scalar or another ansatz or a plain scalar.

        Args:
            other: A scalar or another ansatz.

        Raises:
            TypeError: If other is neither a scalar nor an ansatz.

        Returns:
            PolyExpAnsatz: The product of this ansatz and other.
        """
        if isinstance(other, PolyExpAnsatz):
            new_a = [A1 + A2 for A1, A2 in itertools.product(self.A, other.A)]
            new_b = [b1 + b2 for b1, b2 in itertools.product(self.b, other.b)]
            new_c = [c1 * c2 for c1, c2 in itertools.product(self.c, other.c)]
            return self.__class__(A=new_a, b=new_b, c=new_c)
        else:
            try:
                return self.__class__(self.A, self.b, other * self.c)
            except Exception as e:
                raise TypeError(f"Cannot multiply {self.__class__} and {other.__class__}.") from e

    def __truediv__(self, other: Union[Scalar, PolyExpAnsatz]) -> PolyExpAnsatz:
        r"""Divides this ansatz by a scalar or another ansatz or a plain scalar.

        Args:
            other: A scalar or another ansatz.

        Raises:
            TypeError: If other is neither a scalar nor an ansatz.

        Returns:
            PolyExpAnsatz: The division of this ansatz by other.
        """
        if isinstance(other, PolyExpAnsatz):
            new_a = [A1 - A2 for A1, A2 in itertools.product(self.A, other.A)]
            new_b = [b1 - b2 for b1, b2 in itertools.product(self.b, other.b)]
            new_c = [c1 / c2 for c1, c2 in itertools.product(self.c, other.c)]
            return self.__class__(A=new_a, b=new_b, c=new_c)
        else:
            try:
                return self.__class__(self.A, self.b, self.c / other)
            except Exception as e:
                raise TypeError(f"Cannot divide {self.__class__} and {other.__class__}.") from e

    def __and__(self, other: PolyExpAnsatz) -> PolyExpAnsatz:
        r"""Tensor product of this ansatz with another ansatz.
        Equivalent to :math:`F(a) * G(b)` (with different arguments, that is).
        As it distributes over addition on both self and other,
        the batch size of the result is the product of the batch
        size of this anzatz and the other one.

        Args:
            other: Another ansatz.

        Returns:
            The tensor product of this ansatz and other.
        """
        As = [math.block_diag(a1, a2) for a1 in self.A for a2 in other.A]
        bs = [math.concat([b1, b2], axis=-1) for b1 in self.b for b2 in other.b]
        cs = [math.outer(c1, c2) for c1 in self.c for c2 in other.c]
        return self.__class__(As, bs, cs)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from numba import njit
from numba.cpython.unsafe.tuple import tuple_setitem


@njit
def first_available_pivot(index: tuple[int, ...]) -> tuple[int, tuple[int, ...]]:
    r"""returns the first available pivot for the given index. A pivot is a nearest neighbor
    of the index. Here we pick the first available pivot.

    Arguments:
        index: the index to get the first available pivot of.

    Returns:
        the index that was decremented and the pivot
    """
    for i, v in enumerate(index):
        if v > 0:
            return i, tuple_setitem(index, i, v - 1)
    raise ValueError("Index is zero")


@njit
def smallest_pivot(index: tuple[int, ...]) -> tuple[int, tuple[int, ...]]:
    r"""returns the pivot closest to a zero index. A pivot is a nearest neighbor
    of the index. Here we pick the pivot with the smallest non-zero element.

    Arguments:
        index: the index to get the smallest pivot of.

    Returns:
        (int, tuple) the index of the element that was decremented and the pivot
    """
    min_ = 2**64 - 1
    for i, v in enumerate(index):
        if 0 < v < min_:
            min_ = v
            min_i = i
    if min_ == 2**64 - 1:
        raise ValueError("Index is zero")
    return min_i, tuple_setitem(index, min_i, min_ - 1)



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the tensorflow backend."""

# pylint: disable = missing-function-docstring, missing-class-docstring, wrong-import-position

from typing import Callable, List, Optional, Sequence, Tuple, Union

import os
import numpy as np
import tensorflow_probability as tfp

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
import tensorflow as tf

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "0"


from mrmustard.math.lattice.strategies.compactFock.inputValidation import (
    grad_hermite_multidimensional_1leftoverMode,
    grad_hermite_multidimensional_diagonal,
    hermite_multidimensional_1leftoverMode,
    hermite_multidimensional_diagonal,
    hermite_multidimensional_diagonal_batch,
)

from ..utils.settings import settings
from ..utils.typing import Tensor, Trainable
from .autocast import Autocast
from .backend_base import BackendBase
from .lattice import strategies


# pylint: disable=too-many-public-methods
class BackendTensorflow(BackendBase):  # pragma: no cover
    r"""
    A base class for backends.
    """

    int32 = tf.int32
    float32 = tf.float32
    float64 = tf.float64
    complex64 = tf.complex64
    complex128 = tf.complex128

    def __init__(self):
        super().__init__(name="tensorflow")

    def __repr__(self) -> str:
        return "BackendTensorflow()"

    def abs(self, array: tf.Tensor) -> tf.Tensor:
        return tf.abs(array)

    def any(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.reduce_any(array)

    def arange(self, start: int, limit: int = None, delta: int = 1, dtype=None) -> tf.Tensor:
        dtype = dtype or self.float64
        return tf.range(start, limit, delta, dtype=dtype)

    def asnumpy(self, tensor: tf.Tensor) -> Tensor:
        return np.array(tensor)

    def assign(self, tensor: tf.Tensor, value: tf.Tensor) -> tf.Tensor:
        tensor.assign(value)
        return tensor

    def astensor(self, array: Union[np.ndarray, tf.Tensor], dtype=None) -> tf.Tensor:
        dtype = dtype or np.array(array).dtype.name
        return tf.convert_to_tensor(array, dtype)

    def atleast_1d(self, array: tf.Tensor, dtype=None) -> tf.Tensor:
        return tf.experimental.numpy.atleast_1d(self.astensor(array, dtype))

    def atleast_2d(self, array: tf.Tensor, dtype=None) -> tf.Tensor:
        return tf.experimental.numpy.atleast_2d(self.astensor(array, dtype))

    def atleast_3d(self, array: tf.Tensor, dtype=None) -> tf.Tensor:
        array = self.atleast_2d(self.atleast_1d(array, dtype))
        if len(array.shape) == 2:
            array = self.expand_dims(array, 0)
        return array

    def block_diag(self, mat1: tf.Tensor, mat2: tf.Tensor) -> tf.Tensor:
        Za = self.zeros((mat1.shape[-2], mat2.shape[-1]), dtype=mat1.dtype)
        Zb = self.zeros((mat2.shape[-2], mat1.shape[-1]), dtype=mat1.dtype)
        return self.concat(
            [self.concat([mat1, Za], axis=-1), self.concat([Zb, mat2], axis=-1)],
            axis=-2,
        )

    def block(self, blocks: List[List[tf.Tensor]], axes=(-2, -1)) -> tf.Tensor:
        rows = [self.concat(row, axis=axes[1]) for row in blocks]
        return self.concat(rows, axis=axes[0])

    def boolean_mask(self, tensor: tf.Tensor, mask: tf.Tensor) -> Tensor:
        return tf.boolean_mask(tensor, mask)

    def cast(self, array: tf.Tensor, dtype=None) -> tf.Tensor:
        if dtype is None:
            return array
        return tf.cast(array, dtype)

    def clip(self, array, a_min, a_max) -> tf.Tensor:
        return tf.clip_by_value(array, a_min, a_max)

    def concat(self, values: Sequence[tf.Tensor], axis: int) -> tf.Tensor:
        if any(tf.rank(v) == 0 for v in values):
            return tf.stack(values, axis)
        return tf.concat(values, axis)

    def conj(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.conj(array)

    def constraint_func(
        self, bounds: Tuple[Optional[float], Optional[float]]
    ) -> Optional[Callable]:
        bounds = (
            -np.inf if bounds[0] is None else bounds[0],
            np.inf if bounds[1] is None else bounds[1],
        )
        if bounds != (-np.inf, np.inf):

            def constraint(x):
                return tf.clip_by_value(x, bounds[0], bounds[1])

        else:
            constraint = None
        return constraint

    # pylint: disable=arguments-differ
    @Autocast()
    def convolution(
        self,
        array: tf.Tensor,
        filters: tf.Tensor,
        padding: Optional[str] = None,
        data_format="NWC",
    ) -> tf.Tensor:
        padding = padding or "VALID"
        return tf.nn.convolution(array, filters=filters, padding=padding, data_format=data_format)

    def cos(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.cos(array)

    def cosh(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.cosh(array)

    def det(self, matrix: tf.Tensor) -> tf.Tensor:
        return tf.linalg.det(matrix)

    def diag(self, array: tf.Tensor, k: int = 0) -> tf.Tensor:
        return tf.linalg.diag(array, k=k)

    def diag_part(self, array: tf.Tensor, k: int = 0) -> tf.Tensor:
        return tf.linalg.diag_part(array, k=k)

    def einsum(self, string: str, *tensors) -> tf.Tensor:
        if isinstance(string, str):
            return tf.einsum(string, *tensors)
        return None  # provide same functionality as numpy.einsum or upgrade to opt_einsum

    def exp(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.exp(array)

    def expand_dims(self, array: tf.Tensor, axis: int) -> tf.Tensor:
        return tf.expand_dims(array, axis)

    def expm(self, matrix: tf.Tensor) -> tf.Tensor:
        return tf.linalg.expm(matrix)

    def eye(self, size: int, dtype=None) -> tf.Tensor:
        dtype = dtype or self.float64
        return tf.eye(size, dtype=dtype)

    def eye_like(self, array: tf.Tensor) -> Tensor:
        return tf.eye(array.shape[-1], dtype=array.dtype)

    def from_backend(self, value) -> bool:
        return isinstance(value, (tf.Tensor, tf.Variable))

    def gather(self, array: tf.Tensor, indices: tf.Tensor, axis: int) -> tf.Tensor:
        return tf.gather(array, indices, axis=axis)

    def imag(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.imag(array)

    def inv(self, tensor: tf.Tensor) -> tf.Tensor:
        return tf.linalg.inv(tensor)

    def is_trainable(self, tensor: tf.Tensor) -> bool:
        return isinstance(tensor, tf.Variable)

    def lgamma(self, x: tf.Tensor) -> tf.Tensor:
        return tf.math.lgamma(x)

    def log(self, x: tf.Tensor) -> tf.Tensor:
        return tf.math.log(x)

    @Autocast()
    def matmul(self, *matrices: tf.Tensor) -> tf.Tensor:
        mat = matrices[0]
        for matrix in matrices[1:]:
            mat = tf.matmul(mat, matrix)
        return mat

    @Autocast()
    def matvec(self, a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
        return tf.linalg.matvec(a, b)

    def make_complex(self, real: tf.Tensor, imag: tf.Tensor) -> tf.Tensor:
        return tf.complex(real, imag)

    @Autocast()
    def maximum(self, a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
        return tf.maximum(a, b)

    @Autocast()
    def minimum(self, a: tf.Tensor, b: tf.Tensor) -> tf.Tensor:
        return tf.minimum(a, b)

    def new_variable(
        self,
        value,
        bounds: Union[Tuple[Optional[float], Optional[float]], None],
        name: str,
        dtype=None,
    ):
        bounds = bounds or (None, None)
        dtype = dtype or self.float64
        value = self.astensor(value, dtype)
        return tf.Variable(value, name=name, dtype=dtype, constraint=self.constraint_func(bounds))

    def new_constant(self, value, name: str, dtype=None):
        dtype = dtype or self.float64
        value = self.astensor(value, dtype)
        return tf.constant(value, dtype=dtype, name=name)

    def norm(self, array: tf.Tensor) -> tf.Tensor:
        """Note that the norm preserves the type of array."""
        return tf.linalg.norm(array)

    def ones(self, shape: Sequence[int], dtype=None) -> tf.Tensor:
        dtype = dtype or self.float64
        return tf.ones(shape, dtype=dtype)

    def ones_like(self, array: tf.Tensor) -> tf.Tensor:
        return tf.ones_like(array)

    @Autocast()
    def outer(self, array1: tf.Tensor, array2: tf.Tensor) -> tf.Tensor:
        return tf.tensordot(array1, array2, [[], []])

    def pad(
        self,
        array: tf.Tensor,
        paddings: Sequence[Tuple[int, int]],
        mode="CONSTANT",
        constant_values=0,
    ) -> tf.Tensor:
        return tf.pad(array, paddings, mode, constant_values)

    @staticmethod
    def pinv(matrix: tf.Tensor) -> tf.Tensor:
        return tf.linalg.pinv(matrix)

    @Autocast()
    def pow(self, x: tf.Tensor, y: float) -> tf.Tensor:
        return tf.math.pow(x, y)

    def real(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.real(array)

    def reshape(self, array: tf.Tensor, shape: Sequence[int]) -> tf.Tensor:
        return tf.reshape(array, shape)

    def round(self, array: tf.Tensor, decimals: int = 0) -> tf.Tensor:
        return tf.round(10**decimals * array) / 10**decimals

    def set_diag(self, array: tf.Tensor, diag: tf.Tensor, k: int) -> tf.Tensor:
        return tf.linalg.set_diag(array, diag, k=k)

    def sin(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.sin(array)

    def sinh(self, array: tf.Tensor) -> tf.Tensor:
        return tf.math.sinh(array)

    def solve(self, matrix: tf.Tensor, rhs: tf.Tensor) -> tf.Tensor:
        if len(rhs.shape) == len(matrix.shape) - 1:
            rhs = tf.expand_dims(rhs, -1)
            return tf.linalg.solve(matrix, rhs)[..., 0]
        return tf.linalg.solve(matrix, rhs)

    def sqrt(self, x: tf.Tensor, dtype=None) -> tf.Tensor:
        return tf.sqrt(self.cast(x, dtype))

    def sum(self, array: tf.Tensor, axes: Sequence[int] = None):
        return tf.reduce_sum(array, axes)

    @Autocast()
    def tensordot(self, a: tf.Tensor, b: tf.Tensor, axes: List[int]) -> tf.Tensor:
        return tf.tensordot(a, b, axes)

    def tile(self, array: tf.Tensor, repeats: Sequence[int]) -> tf.Tensor:
        return tf.tile(array, repeats)

    def trace(self, array: tf.Tensor, dtype=None) -> tf.Tensor:
        return self.cast(tf.linalg.trace(array), dtype)

    def transpose(self, a: tf.Tensor, perm: Sequence[int] = None) -> tf.Tensor:
        if a is None:
            return None  # TODO: remove and address None inputs where tranpose is used
        return tf.transpose(a, perm)

    @Autocast()
    def update_tensor(self, tensor: tf.Tensor, indices: tf.Tensor, values: tf.Tensor) -> tf.Tensor:
        return tf.tensor_scatter_nd_update(tensor, indices, values)

    @Autocast()
    def update_add_tensor(
        self, tensor: tf.Tensor, indices: tf.Tensor, values: tf.Tensor
    ) -> tf.Tensor:
        return tf.tensor_scatter_nd_add(tensor, indices, values)

    def zeros(self, shape: Sequence[int], dtype=None) -> tf.Tensor:
        dtype = dtype or self.float64
        return tf.zeros(shape, dtype=dtype)

    def zeros_like(self, array: tf.Tensor) -> tf.Tensor:
        return tf.zeros_like(array)

    def map_fn(self, func, elements):
        return tf.map_fn(func, elements)

    def squeeze(self, tensor, axis=None):
        return tf.squeeze(tensor, axis=axis or [])

    def cholesky(self, input: Tensor):
        return tf.linalg.cholesky(input)

    def Categorical(self, probs: Tensor, name: str):
        return tfp.distributions.Categorical(probs=probs, name=name)

    def MultivariateNormalTriL(self, loc: Tensor, scale_tril: Tensor):
        return tfp.distributions.MultivariateNormalTriL(loc=loc, scale_tril=scale_tril)

    @staticmethod
    def eigh(tensor: tf.Tensor) -> Tensor:
        return tf.linalg.eigh(tensor)

    @staticmethod
    def eigvals(tensor: tf.Tensor) -> Tensor:
        return tf.linalg.eigvals(tensor)

    @staticmethod
    def xlogy(x: tf.Tensor, y: tf.Tensor) -> Tensor:
        return tf.math.xlogy(x, y)

    def sqrtm(self, tensor: tf.Tensor, dtype, rtol=1e-05, atol=1e-08) -> Tensor:
        # The sqrtm function has issues with matrices that are close to zero, hence we branch
        if np.allclose(tensor, 0, rtol=rtol, atol=atol):
            ret = self.zeros_like(tensor)
        else:
            ret = tf.linalg.sqrtm(tensor)

        if dtype is None:
            return self.cast(ret, self.complex128)
        return self.cast(ret, dtype)

    # ~~~~~~~~~~~~~~~~~
    # Special functions
    # ~~~~~~~~~~~~~~~~~

    def DefaultEuclideanOptimizer(self) -> tf.keras.optimizers.legacy.Optimizer:
        return tf.keras.optimizers.legacy.Adam(learning_rate=0.001)

    def value_and_gradients(
        self, cost_fn: Callable, parameters: List[Trainable]
    ) -> Tuple[tf.Tensor, List[tf.Tensor]]:
        r"""Computes the loss and gradients of the given cost function.

        Args:
            cost_fn (Callable with no args): The cost function.
            parameters (List[Trainable]): The parameters to optimize.

        Returns:
            tuple(Tensor, List[Tensor]): the loss and the gradients
        """
        with tf.GradientTape() as tape:
            loss = cost_fn()
        gradients = tape.gradient(loss, parameters)
        return loss, gradients

    @tf.custom_gradient
    def hermite_renormalized(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, shape: Tuple[int]
    ) -> Tuple[tf.Tensor, Callable]:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx + 1/2*Ax^2)` at zero, where the series has :math:`sqrt(n!)`
        at the denominator rather than :math:`n!`. It computes all the amplitudes within the
        tensor of given shape.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            shape: The shape of the final tensor.

        Returns:
            The renormalized Hermite polynomial of given shape.
        """

        precision_bits = settings.PRECISION_BITS_HERMITE_POLY

        A, B, C = self.asnumpy(A), self.asnumpy(B), self.asnumpy(C)

        if precision_bits == 128:  # numba
            G = strategies.vanilla(tuple(shape), A, B, C)
        else:  # julia
            # The following import must come after running "jl = Julia(compiled_modules=False)" in settings.py
            from julia import Main as Main_julia  # pylint: disable=import-outside-toplevel

            A, B, C = (
                A.astype(np.complex128),
                B.astype(np.complex128),
                C.astype(np.complex128),
            )

            G = Main_julia.Vanilla.vanilla(
                A, B, C.item(), np.array(shape, dtype=np.int64), precision_bits
            )

        def grad(dLdGconj):
            dLdA, dLdB, dLdC = strategies.vanilla_vjp(G, C, np.conj(dLdGconj))
            return self.conj(dLdA), self.conj(dLdB), self.conj(dLdC)

        return G, grad

    def hermite_renormalized_batch(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, shape: Tuple[int]
    ) -> tf.Tensor:
        _A, _B, _C = self.asnumpy(A), self.asnumpy(B), self.asnumpy(C)

        G = strategies.vanilla_batch(tuple(shape), _A, _B, _C)
        return G

    @tf.custom_gradient
    def hermite_renormalized_binomial(
        self,
        A: tf.Tensor,
        B: tf.Tensor,
        C: tf.Tensor,
        shape: Tuple[int],
        max_l2: Optional[float],
        global_cutoff: Optional[int],
    ) -> tf.Tensor:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx + 1/2*Ax^2)` at zero, where the series has :math:`sqrt(n!)`
        at the denominator rather than :math:`n!`. The computation fills a tensor of given shape
        up to a given L2 norm or global cutoff, whichever applies first. The max_l2 value, if
        not provided, is set to the default value of the AUTOCUTOFF_PROBABILITY setting.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            shape: The shape of the final tensor (local cutoffs).
            max_l2 (float): The maximum squared L2 norm of the tensor.
            global_cutoff (optional int): The global cutoff.

        Returns:
            The renormalized Hermite polynomial of given shape.
        """
        _A, _B, _C = self.asnumpy(A), self.asnumpy(B), self.asnumpy(C)
        G, _ = strategies.binomial(
            tuple(shape),
            _A,
            _B,
            _C,
            max_l2=max_l2 or settings.AUTOCUTOFF_PROBABILITY,
            global_cutoff=global_cutoff or sum(shape) - len(shape) + 1,
        )

        def grad(dLdGconj):
            dLdA, dLdB, dLdC = strategies.vanilla_vjp(G, _C, np.conj(dLdGconj))
            return self.conj(dLdA), self.conj(dLdB), self.conj(dLdC)

        return G, grad

    def reorder_AB_bargmann(self, A: tf.Tensor, B: tf.Tensor) -> Tuple[tf.Tensor, tf.Tensor]:
        r"""In mrmustard.math.compactFock.compactFock~ dimensions of the Fock representation are ordered like [mode0,mode0,mode1,mode1,...]
        while in mrmustard.physics.bargmann the ordering is [mode0,mode1,...,mode0,mode1,...]. Here we reorder A and B.
        """
        ordering = (
            np.arange(2 * A.shape[0] // 2).reshape(2, -1).T.flatten()
        )  # ordering is [0,2,4,...,1,3,5,...]
        A = tf.gather(A, ordering, axis=1)
        A = tf.gather(A, ordering)
        B = tf.gather(B, ordering, axis=0)
        return A, B

    def hermite_renormalized_diagonal(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, cutoffs: Tuple[int]
    ) -> tf.Tensor:
        A, B = self.reorder_AB_bargmann(A, B)
        return self.hermite_renormalized_diagonal_reorderedAB(A, B, C, cutoffs=cutoffs)

    @tf.custom_gradient
    def hermite_renormalized_diagonal_reorderedAB(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, cutoffs: Tuple[int]
    ) -> tf.Tensor:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx - Ax^2)` at zero, where the series has :math:`sqrt(n!)` at the
        denominator rather than :math:`n!`. Note the minus sign in front of ``A``.

        Calculates the diagonal of the Fock representation (i.e. the PNR detection probabilities of all modes)
        by applying the recursion relation in a selective manner.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            cutoffs: upper boundary of photon numbers in each mode

        Returns:
            The renormalized Hermite polynomial.
        """
        A, B, C = self.asnumpy(A), self.asnumpy(B), self.asnumpy(C)
        precision_bits = settings.PRECISION_BITS_HERMITE_POLY

        if precision_bits == 128:  # numba (complex128)
            poly0, poly2, poly1010, poly1001, poly1 = tf.numpy_function(
                hermite_multidimensional_diagonal, [A, B, C, cutoffs], [A.dtype] * 5
            )
        else:  # julia (higher precision than complex128)
            # The following import must come after running "jl = Julia(compiled_modules=False)" in settings.py
            from julia import Main as Main_julia  # pylint: disable=import-outside-toplevel

            (
                poly0,
                poly2,
                poly1010,
                poly1001,
                poly1,
            ) = Main_julia.DiagonalAmps.fock_diagonal_amps(
                A, B, C.item(), tuple(cutoffs), precision_bits
            )

        def grad(dLdpoly):
            if precision_bits == 128:  # numba (complex128)
                dpoly_dC, dpoly_dA, dpoly_dB = tf.numpy_function(
                    grad_hermite_multidimensional_diagonal,
                    [A, B, C.item(), poly0, poly2, poly1010, poly1001, poly1],
                    [poly0.dtype] * 3,
                )
            else:  # julia (higher precision than complex128)
                dpoly_dC = poly0 / C.item()
                dpoly_dA, dpoly_dB = Main_julia.DiagonalGrad.fock_diagonal_grad(
                    A, B, poly0, poly2, poly1010, poly1001, poly1, precision_bits
                )

            ax = tuple(range(dLdpoly.ndim))
            dLdA = self.sum(dLdpoly[..., None, None] * self.conj(dpoly_dA), axes=ax)
            dLdB = self.sum(dLdpoly[..., None] * self.conj(dpoly_dB), axes=ax)
            dLdC = self.sum(dLdpoly * self.conj(dpoly_dC), axes=ax)
            return dLdA, dLdB, dLdC

        return poly0, grad

    def hermite_renormalized_diagonal_batch(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, cutoffs: Tuple[int]
    ) -> tf.Tensor:
        r"""Same as hermite_renormalized_diagonal but works for a batch of different B's."""
        A, B = self.reorder_AB_bargmann(A, B)
        return self.hermite_renormalized_diagonal_reorderedAB_batch(A, B, C, cutoffs=cutoffs)

    def hermite_renormalized_diagonal_reorderedAB_batch(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, cutoffs: Tuple[int]
    ) -> tf.Tensor:
        r"""Same as hermite_renormalized_diagonal_reorderedAB but works for a batch of different B's.

        Args:
            A: The A matrix.
            B: The B vectors.
            C: The C scalar.
            cutoffs: upper boundary of photon numbers in each mode

        Returns:
            The renormalized Hermite polynomial from different B values.
        """
        A, B, C = self.asnumpy(A), self.asnumpy(B), self.asnumpy(C)

        poly0, _, _, _, _ = tf.numpy_function(
            hermite_multidimensional_diagonal_batch, [A, B, C, cutoffs], [A.dtype] * 5
        )

        return poly0

    def hermite_renormalized_1leftoverMode(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, cutoffs: Tuple[int]
    ) -> tf.Tensor:
        r"""First, reorder A and B parameters of Bargmann representation to match conventions in mrmustard.math.compactFock.compactFock~
        Then, calculate the required renormalized multidimensional Hermite polynomial.
        """
        A, B = self.reorder_AB_bargmann(A, B)
        return self.hermite_renormalized_1leftoverMode_reorderedAB(A, B, C, cutoffs=cutoffs)

    @tf.custom_gradient
    def hermite_renormalized_1leftoverMode_reorderedAB(
        self, A: tf.Tensor, B: tf.Tensor, C: tf.Tensor, cutoffs: Tuple[int]
    ) -> tf.Tensor:
        r"""Renormalized multidimensional Hermite polynomial given by the "exponential" Taylor
        series of :math:`exp(C + Bx - Ax^2)` at zero, where the series has :math:`sqrt(n!)` at the
        denominator rather than :math:`n!`. Note the minus sign in front of ``A``.

        Calculates all possible Fock representations of mode 0,
        where all other modes are PNR detected.
        This is done by applying the recursion relation in a selective manner.

        Args:
            A: The A matrix.
            B: The B vector.
            C: The C scalar.
            cutoffs: upper boundary of photon numbers in each mode

        Returns:
            The renormalized Hermite polynomial.
        """
        A, B, C = self.asnumpy(A), self.asnumpy(B), self.asnumpy(C)
        precision_bits = settings.PRECISION_BITS_HERMITE_POLY

        if precision_bits == 128:  # numba (complex128)
            poly0, poly2, poly1010, poly1001, poly1 = tf.numpy_function(
                hermite_multidimensional_1leftoverMode,
                [A, B, C.item(), cutoffs],
                [A.dtype] * 5,
            )
        else:  # julia (higher precision than complex128)
            # The following import must come after running "jl = Julia(compiled_modules=False)" in settings.py
            from julia import Main as Main_julia  # pylint: disable=import-outside-toplevel

            (
                poly0,
                poly2,
                poly1010,
                poly1001,
                poly1,
            ) = Main_julia.LeftoverModeAmps.fock_1leftoverMode_amps(
                A, B, C.item(), tuple(cutoffs), precision_bits
            )

        def grad(dLdpoly):
            if precision_bits == 128:  # numba (complex128)
                dpoly_dC, dpoly_dA, dpoly_dB = tf.numpy_function(
                    grad_hermite_multidimensional_1leftoverMode,
                    [A, B, C, poly0, poly2, poly1010, poly1001, poly1],
                    [poly0.dtype] * 3,
                )
            else:  # julia (higher precision than complex128)
                dpoly_dC = poly0 / C.item()
                (
                    dpoly_dA,
                    dpoly_dB,
                ) = Main_julia.LeftoverModeGrad.fock_1leftoverMode_grad(
                    A, B, poly0, poly2, poly1010, poly1001, poly1, precision_bits
                )

            ax = tuple(range(dLdpoly.ndim))
            dLdA = self.sum(dLdpoly[..., None, None] * self.conj(dpoly_dA), axes=ax)
            dLdB = self.sum(dLdpoly[..., None] * self.conj(dpoly_dB), axes=ax)
            dLdC = self.sum(dLdpoly * self.conj(dpoly_dC), axes=ax)
            return dLdA, dLdB, dLdC

        return poly0, grad

    @tf.custom_gradient
    def getitem(tensor, *, key):
        """A differentiable pure equivalent of numpy's ``value = tensor[key]``."""
        value = np.array(tensor)[key]

        def grad(dy):
            dL_dtensor = np.zeros_like(tensor)
            dL_dtensor[key] = dy
            return dL_dtensor

        return value, grad

    @tf.custom_gradient
    def setitem(tensor, value, *, key):
        """A differentiable pure equivalent of numpy's ``tensor[key] = value``."""
        _tensor = np.array(tensor)
        value = np.array(value)
        _tensor[key] = value

        def grad(dy):
            dL_dtensor = np.array(dy)
            dL_dtensor[key] = 0.0
            # unbroadcasting the gradient
            implicit_broadcast = list(range(_tensor.ndim - value.ndim))
            explicit_broadcast = [
                _tensor.ndim - value.ndim + j for j in range(value.ndim) if value.shape[j] == 1
            ]
            dL_dvalue = np.sum(
                np.array(dy)[key], axis=tuple(implicit_broadcast + explicit_broadcast)
            )
            dL_dvalue = np.expand_dims(
                dL_dvalue, [i - len(implicit_broadcast) for i in explicit_broadcast]
            )
            return dL_dtensor, dL_dvalue

        return _tensor, grad



# Copyright 2021 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the implementation of the :class:`Transformation` class."""


# pylint: disable = missing-function-docstring

from __future__ import annotations

from typing import Callable, Iterable, Optional, Sequence, Tuple, Union

import numpy as np

from mrmustard import math, settings
from mrmustard.math.parameter_set import ParameterSet
from mrmustard.math.parameters import Constant, Variable
from mrmustard.math.tensor_networks import Tensor
from mrmustard.physics import bargmann, fock, gaussian
from mrmustard.utils.typing import RealMatrix, RealVector

from .state import State


class Transformation(Tensor):
    r"""
    Base class for all Transformations.
    """

    def __init__(
        self,
        name: str,
        modes_in_ket: Optional[list[int]] = None,
        modes_out_ket: Optional[list[int]] = None,
        modes_in_bra: Optional[list[int]] = None,
        modes_out_bra: Optional[list[int]] = None,
    ):
        super().__init__(
            name=name,
            modes_in_ket=modes_in_ket,
            modes_out_ket=modes_out_ket,
            modes_in_bra=modes_in_bra,
            modes_out_bra=modes_out_bra,
        )
        self._parameter_set = ParameterSet()

    def _add_parameter(self, parameter: Union[Constant, Variable]):
        r"""
        Adds a parameter to a transformation.

        Args:
            parameter: The parameter to add.
        """
        self.parameter_set.add_parameter(parameter)
        self.__dict__[parameter.name] = parameter

    @property
    def parameter_set(self):
        r"""
        The set of parameters for this transformation.
        """
        return self._parameter_set

    def primal(self, state: State) -> State:
        r"""Applies this transformation to the given ``state`` and returns the transformed state.

        Args:
            state (State): the state to transform

        Returns:
            State: the transformed state
        """
        if state.is_gaussian:
            new_state = self._transform_gaussian(state, dual=False)
        else:
            new_state = self._transform_fock(state, dual=False)
        return new_state

    def dual(self, state: State) -> State:
        r"""Applies the dual of this transformation to the given ``state`` and returns the transformed state.

        Args:
            state (State): the state to transform

        Returns:
            State: the transformed state
        """
        if state.is_gaussian:
            new_state = self._transform_gaussian(state, dual=True)
        else:
            new_state = self._transform_fock(state, dual=True)
        return new_state

    def _transform_gaussian(self, state: State, dual: bool) -> State:
        r"""Transforms a Gaussian state into a Gaussian state.

        Args:
            state (State): the state to transform
            dual (bool): whether to apply the dual channel

        Returns:
            State: the transformed state
        """
        X, Y, d = self.XYd(allow_none=False) if not dual else self.XYd_dual(allow_none=False)
        cov, means = gaussian.CPTP(state.cov, state.means, X, Y, d, state.modes, self.modes)
        new_state = State(
            cov=cov, means=means, modes=state.modes, _norm=state.norm
        )  # NOTE: assumes modes don't change
        return new_state

    @property
    def num_modes(self) -> int:
        r"""The number of modes on which the transformation acts."""
        return len(self.modes)

    def _validate_modes(self, modes):
        pass

    @property
    def X_matrix(self) -> Optional[RealMatrix]:
        return None

    @property
    def Y_matrix(self) -> Optional[RealMatrix]:
        return None

    @property
    def d_vector(self) -> Optional[RealVector]:
        return None

    @property
    def X_matrix_dual(self) -> Optional[RealMatrix]:
        if (X := self.X_matrix) is None:
            return None
        return gaussian.math.inv(X)

    @property
    def Y_matrix_dual(self) -> Optional[RealMatrix]:
        if (Y := self.Y_matrix) is None:
            return None
        if (Xdual := self.X_matrix_dual) is None:
            return Y
        return math.matmul(math.matmul(Xdual, Y), math.transpose(Xdual))

    @property
    def d_vector_dual(self) -> Optional[RealVector]:
        if (d := self.d_vector) is None:
            return None
        if (Xdual := self.X_matrix_dual) is None:
            return -d
        return -math.matmul(Xdual, d)

    def bargmann(self, numpy=False):
        X, Y, d = self.XYd(allow_none=False)
        if self.is_unitary:
            A, B, C = bargmann.wigner_to_bargmann_U(X, d)
        else:
            A, B, C = bargmann.wigner_to_bargmann_Choi(X, Y, d)
        if numpy:
            return math.asnumpy(A), math.asnumpy(B), math.asnumpy(C)
        return A, B, C

    def choi(
        self,
        cutoffs: Optional[Sequence[int]] = None,
        shape: Optional[Sequence[int]] = None,
        dual: bool = False,
    ):
        r"""Returns the Choi representation of the transformation.

        If specified, ``shape`` takes precedence over ``cutoffs``.
        The ``shape`` is in the order ``(out_L, in_L, out_R, in_R)``.

        Args:
            cutoffs: the cutoffs of the input and output modes
            shape: the shape of the Choi matrix
            dual: whether to return the dual Choi
        """
        N = self.num_modes
        if cutoffs is None:
            pass
        elif len(cutoffs) != N:
            raise ValueError(f"len(cutoffs) must be {self.num_modes} (got {len(cutoffs)})")

        shape = shape or tuple(cutoffs) * 4

        if self.is_unitary:
            shape = shape[: 2 * self.num_modes]
            U = self.U(shape[: self.num_modes])
            Udual = self.U(shape[self.num_modes :])
            if dual:
                return fock.U_to_choi(U=Udual, Udual=U)
            return fock.U_to_choi(U=U, Udual=Udual)

        X, Y, d = self.XYd(allow_none=False)
        choi = fock.wigner_to_fock_Choi(X, Y, d, shape=shape)
        if dual:
            n = len(shape) // 4
            N0 = list(range(0, n))
            N1 = list(range(n, 2 * n))
            N2 = list(range(2 * n, 3 * n))
            N3 = list(range(3 * n, 4 * n))
            choi = math.conj(math.transpose(choi, N1 + N0 + N3 + N2))  # if dual we flip out-in
        return choi

    def XYd(
        self, allow_none: bool = True
    ) -> Tuple[Optional[RealMatrix], Optional[RealMatrix], Optional[RealVector]]:
        r"""Returns the ```(X, Y, d)``` triple.

        Override in subclasses if computing ``X``, ``Y`` and ``d`` together is more efficient.
        """
        if allow_none:
            return self.X_matrix, self.Y_matrix, self.d_vector
        X = math.eye(2 * self.num_modes) if self.X_matrix is None else self.X_matrix
        Y = math.zeros_like(X) if self.Y_matrix is None else self.Y_matrix
        d = math.zeros_like(X[:, 0]) if self.d_vector is None else self.d_vector
        return X, Y, d

    def XYd_dual(
        self, allow_none: bool = True
    ) -> tuple[Optional[RealMatrix], Optional[RealMatrix], Optional[RealVector]]:
        r"""Returns the ```(X, Y, d)``` triple of the dual of the current transformation.

        Override in subclasses if computing ``Xdual``, ``Ydual`` and ``ddual`` together is more efficient.
        """
        if allow_none:
            return self.X_matrix_dual, self.Y_matrix_dual, self.d_vector_dual
        Xdual = math.eye(2 * self.num_modes) if self.X_matrix_dual is None else self.X_matrix_dual
        Ydual = math.zeros_like(Xdual) if self.Y_matrix_dual is None else self.Y_matrix_dual
        ddual = math.zeros_like(Xdual[:, 0]) if self.d_vector_dual is None else self.d_vector_dual
        return Xdual, Ydual, ddual

    def __getitem__(self, items) -> Callable:
        r"""Sets the modes on which the transformation acts.

        Allows transformations to be used as: ``output = transf[0,1](input)``,  e.g. acting on
        modes 0 and 1.
        """
        #  TODO: this won't work when we want to reuse the same op for different modes in a circuit.
        # i.e. `psi = op[0](psi); psi = op[1](psi)` is ok, but `circ = Circuit([op[0], op[1]])` won't work.
        if isinstance(items, int):
            modes = [items]
        elif isinstance(items, slice):
            modes = list(range(items.start, items.stop, items.step))
        elif isinstance(items, (Sequence, Iterable)):
            modes = list(items)
        else:
            raise ValueError(f"{items} is not a valid slice or list of modes.")
        if self.is_unitary:
            self.change_modes(modes, modes)
        else:
            self.change_modes(modes, modes, modes, modes)
        return self

    def __rshift__(self, other: Transformation):
        r"""Concatenates self with other (other after self).

        If any of the two is a circuit, all the ops in it migrate to the new circuit that is returned.
        E.g., ``circ = Sgate(1.0)[0,1] >> Dgate(0.2)[0] >> BSgate(np.pi/4)[0,1]``

        Args:
            other: another transformation

        Returns:
            Circuit: A circuit that concatenates self with other
        """
        from mrmustard.lab.circuit import Circuit  # pylint: disable=import-outside-toplevel

        ops1 = self._ops if isinstance(self, Circuit) else [self]
        ops2 = other._ops if isinstance(other, Circuit) else [other]
        return Circuit(ops1 + ops2)

    def __lshift__(self, other: Union[State, Transformation]):
        r"""Applies the dual of self to other.

        If other is a state, the dual of self is applied to the state.
        If other is a transformation, the dual of self is concatenated after other (in the dual sense).

        E.g.
        .. code-block::

            Sgate(0.1) << Coherent(0.5)   # state
            Sgate(0.1) << Dgate(0.2)      # transformation

        Args:
            other: a state or a transformation

        Returns:
            State: the state transformed via the dual transformation or the transformation
            concatenated after other
        """
        if isinstance(other, State):
            return self.dual(other)
        if isinstance(other, Transformation):
            return self >> other  # so that the dual is self.dual(other.dual(x))
        raise ValueError(
            f"{other} of type {other.__class__} is not a valid state or transformation."
        )

    def __repr__(self):
        class_name = self.name
        modes = self.modes

        parameters = {k: v for k, v in self.__dict__.items() if isinstance(v, (Constant, Variable))}
        param_str_rep = [
            f"{name}={repr(math.asnumpy(par.value))}" for name, par in parameters.items()
        ]

        params_str = ", ".join(sorted(param_str_rep))

        return f"{class_name}({params_str}, modes = {modes})".replace("\n", "")

    def __str__(self):
        class_name = self.name
        modes = self.modes
        return f"<{class_name} object at {hex(id(self))} acting on modes {modes}>"

    def _repr_markdown_(self):
        header = (
            f"##### {self.__class__.__qualname__} on modes {self.modes}\n"
            "|Parameters|dtype|Value|Bounds|Shape|Trainable|\n"
            "| :-:      | :-: | :-: | :-:  | :-: | :-:     |\n"
        )

        body = ""
        with np.printoptions(precision=6, suppress=True):
            parameters = {
                k: v for k, v in self.__dict__.items() if isinstance(v, (Constant, Variable))
            }
            for name, par in parameters.items():
                par_value = repr(math.asnumpy(par.value)).replace("\n", "<br>")
                body += (
                    f"| {name}"
                    f"| {par.value.dtype.name}"
                    f"| {par_value}"
                    f"| {str(getattr(par.value, 'bounds', 'None'))}"
                    f"| {par.value.shape}"
                    f"| {str(math.is_trainable(par.value))}"
                    "|\n"
                )

        return header + body


class Unitary(Transformation):
    r"""
    A unitary transformation.

    Args:
        name: The name of this unitary.
        modes: The modes that this unitary acts on.
    """

    def __init__(self, name: str, modes: list[int]):
        super().__init__(name=name, modes_in_ket=modes, modes_out_ket=modes)
        self.is_unitary = True

    def value(self, shape: Tuple[int]):
        return self.U(shape=shape)

    def _transform_fock(self, state: State, dual=False) -> State:
        op_idx = [state.modes.index(m) for m in self.modes]
        U = self.U(cutoffs=[state.cutoffs[i] for i in op_idx])
        if state.is_hilbert_vector:
            return State(ket=fock.apply_kraus_to_ket(U, state.ket(), op_idx), modes=state.modes)
        return State(dm=fock.apply_kraus_to_dm(U, state.dm(), op_idx), modes=state.modes)

    def U(
        self,
        cutoffs: Optional[Sequence[int]] = None,
        shape: Optional[Sequence[int]] = None,
    ):
        r"""Returns the unitary representation of the transformation.

        If specified, ``shape`` takes precedence over ``cutoffs``.
        ``shape`` is in the order ``(out, in)``.

        Note that for a unitary transformation on N modes, ``len(cutoffs)`` is ``N``
        and ``len(shape)`` is ``2N``.

        Args:
            cutoffs: the cutoffs of the input and output modes
            shape: the shape of the unitary matrix

        Returns:
            ComplexTensor: the unitary matrix in Fock representation
        """
        if cutoffs is None:
            pass
        elif len(cutoffs) != self.num_modes:
            raise ValueError(f"len(cutoffs) must be {self.num_modes} (got {len(cutoffs)})")
        shape = shape or tuple(cutoffs) * 2
        X, _, d = self.XYd(allow_none=False)
        return fock.wigner_to_fock_U(X, d, shape=shape)

    def __eq__(self, other):
        r"""Returns ``True`` if the two transformations are equal."""
        if not isinstance(other, Unitary):
            return False
        if not (self.is_gaussian and other.is_gaussian):
            return np.allclose(
                self.U(cutoffs=[settings.EQ_TRANSFORMATION_CUTOFF] * 2 * self.num_modes),
                other.U(cutoffs=[settings.EQ_TRANSFORMATION_CUTOFF] * 2 * self.num_modes),
                rtol=settings.EQ_TRANSFORMATION_RTOL_FOCK,
            )
        sX, sY, sd = self.XYd(allow_none=False)
        oX, oY, od = other.XYd(allow_none=False)
        assert np.isclose(np.linalg.norm(sY), 0)
        assert np.isclose(np.linalg.norm(oY), 0)
        return np.allclose(sX, oX) and np.allclose(sd, od)


class Channel(Transformation):
    r"""
    A quantum channel.

    Args:
        name: The name of this channel.
        modes: The modes that this channel acts on.
    """

    def __init__(self, name: str, modes: list[int]):
        super().__init__(
            name=name,
            modes_in_ket=modes,
            modes_out_ket=modes,
            modes_in_bra=modes,
            modes_out_bra=modes,
        )
        self.is_unitary = False

    def _transform_fock(self, state: State, dual: bool = False) -> State:
        op_idx = [state.modes.index(m) for m in self.modes]
        choi = self.choi(cutoffs=[state.cutoffs[i] for i in op_idx], dual=dual)
        if state.is_hilbert_vector:
            return State(dm=fock.apply_choi_to_ket(choi, state.ket(), op_idx), modes=state.modes)
        return State(dm=fock.apply_choi_to_dm(choi, state.dm(), op_idx), modes=state.modes)

    def value(self, shape: Tuple[int]):
        return self.choi(shape=shape)

    def __eq__(self, other):
        r"""Returns ``True`` if the two transformations are equal."""
        if not isinstance(other, Channel):
            return False
        if not (self.is_gaussian and other.is_gaussian):
            return np.allclose(
                self.choi(cutoffs=[settings.EQ_TRANSFORMATION_CUTOFF] * 4 * self.num_modes),
                other.choi(cutoffs=[settings.EQ_TRANSFORMATION_CUTOFF] * 4 * self.num_modes),
                rtol=settings.EQ_TRANSFORMATION_RTOL_FOCK,
            )
        sX, sY, sd = self.XYd(allow_none=False)
        oX, oY, od = other.XYd(allow_none=False)
        return np.allclose(sX, oX) and np.allclose(sY, oY) and np.allclose(sd, od)



# Copyright 2023 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains functions for transforming to the Bargmann representation.
"""
from typing import Sequence, Tuple

import numpy as np

from mrmustard import math, settings
from mrmustard.physics.husimi import pq_to_aadag, wigner_to_husimi
from mrmustard.utils.typing import ComplexMatrix, ComplexVector


def cayley(X, c):
    r"""Returns the Cayley transform of a matrix:
    :math:`cay(X) = (X - cI)(X + cI)^{-1}`

    Args:
        c (float): the parameter of the Cayley transform
        X (Tensor): a matrix

    Returns:
        Tensor: the Cayley transform of X
    """
    I = math.eye(X.shape[0], dtype=X.dtype)
    return math.solve(X + c * I, X - c * I)


def wigner_to_bargmann_rho(cov, means):
    r"""Converts the wigner representation in terms of covariance matrix and mean vector into the Bargmann `A,B,C` triple
    for a density matrix (i.e. for `M` modes, `A` has shape `2M x 2M` and `B` has shape `2M`).
    The order of the rows/columns of A and B corresponds to a density matrix with the usual ordering of the indices.

    Note that here A and B are defined with respect to the literature.
    """
    N = cov.shape[-1] // 2
    A = math.matmul(math.Xmat(N), cayley(pq_to_aadag(cov), c=0.5))
    Q, beta = wigner_to_husimi(cov, means)
    b = math.solve(Q, beta)
    B = math.conj(b)
    num_C = math.exp(-0.5 * math.sum(math.conj(beta) * b))
    den_C = math.sqrt(math.det(Q), dtype=num_C.dtype)
    C = num_C / den_C
    return A, B, C


def wigner_to_bargmann_psi(cov, means):
    r"""Converts the wigner representation in terms of covariance matrix and mean vector into the Bargmann A,B,C triple
    for a Hilbert vector (i.e. for M modes, A has shape M x M and B has shape M).
    """
    N = cov.shape[-1] // 2
    A, B, C = wigner_to_bargmann_rho(cov, means)
    return A[N:, N:], B[N:], math.sqrt(C)
    # NOTE: c for th psi is to calculated from the global phase formula.


def wigner_to_bargmann_Choi(X, Y, d):
    r"""Converts the wigner representation in terms of covariance matrix and mean vector into the Bargmann `A,B,C` triple
    for a channel (i.e. for M modes, A has shape 4M x 4M and B has shape 4M)."""
    N = X.shape[-1] // 2
    I2 = math.eye(2 * N, dtype=X.dtype)
    XT = math.transpose(X)
    xi = 0.5 * (I2 + math.matmul(X, XT) + 2 * Y / settings.HBAR)
    xi_inv = math.inv(xi)
    A = math.block(
        [
            [I2 - xi_inv, math.matmul(xi_inv, X)],
            [math.matmul(XT, xi_inv), I2 - math.matmul(math.matmul(XT, xi_inv), X)],
        ]
    )
    I = math.eye(N, dtype="complex128")
    o = math.zeros_like(I)
    R = math.block(
        [[I, 1j * I, o, o], [o, o, I, -1j * I], [I, -1j * I, o, o], [o, o, I, 1j * I]]
    ) / np.sqrt(2)
    A = math.matmul(math.matmul(R, A), math.dagger(R))
    A = math.matmul(math.Xmat(2 * N), A)
    b = math.matvec(xi_inv, d)
    B = math.matvec(math.conj(R), math.concat([b, -math.matvec(XT, b)], axis=-1)) / math.sqrt(
        settings.HBAR, dtype=R.dtype
    )
    C = math.exp(-0.5 * math.sum(d * b) / settings.HBAR) / math.sqrt(math.det(xi), dtype=b.dtype)
    # now A and B have order [out_r, in_r out_l, in_l].
    return A, B, math.cast(C, "complex128")


def wigner_to_bargmann_U(X, d):
    r"""Converts the wigner representation in terms of covariance matrix and mean vector into the Bargmann `A,B,C` triple
    for a unitary (i.e. for `M` modes, `A` has shape `2M x 2M` and `B` has shape `2M`).
    """
    N = X.shape[-1] // 2
    A, B, C = wigner_to_bargmann_Choi(X, math.zeros_like(X), d)
    return A[2 * N :, 2 * N :], B[2 * N :], math.sqrt(C)


def complex_gaussian_integral(
    Abc: tuple, idx_z: tuple[int, ...], idx_zconj: tuple[int, ...], measure: float = -1
):
    r"""Computes the Gaussian integral of the exponential of a complex quadratic form.
    The integral is defined as (note that in general we integrate over a subset of 2m dimensions):

    :math:`\int_{C^m} F(z) d\mu(z)`

    where

    :math:`F(z) = \textrm{exp}(-0.5 z^T A z + b^T z)`

    Here, ``z`` is an ``n``-dim complex vector, ``A`` is an ``n x n`` complex matrix,
    ``b`` is an ``n``-dim complex vector, ``c`` is a complex scalar, and :math:`d\mu(z)`
    is a non-holomorphic complex measure over a subset of m pairs of z,z* variables. These
    are specified by the indices ``idx_z`` and ``idx_zconj``. The ``measure`` parameter is
    the exponent of the measure:

    :math: `dmu(z) = \textrm{exp}(m * |z|^2) \frac{d^{2n}z}{\pi^n} = \frac{1}{\pi^n}\textrm{exp}(m * |z|^2) d\textrm{Re}(z) d\textrm{Im}(z)`

    Arguments:
        A,b,c: the ``(A,b,c)`` triple
        idx_z: the tuple of indices of the z variables
        idx_zconj: the tuple of indices of the z* variables
        measure: the exponent of the measure (default is -1: Bargmann measure)

    Returns:
        The ``(A,b,c)`` triple of the result of the integral
    """
    A, b, c = Abc
    if len(idx_z) != len(idx_zconj):
        raise ValueError("idx_z and idx_zconj must have the same length")
    n = len(idx_z)
    idx = tuple(idx_z) + tuple(idx_zconj)
    not_idx = tuple(i for i in range(A.shape[-1]) if i not in idx)

    I = math.eye(n, dtype=A.dtype)
    Z = math.zeros((n, n), dtype=A.dtype)
    X = math.block([[Z, I], [I, Z]])
    M = math.gather(math.gather(A, idx, axis=-1), idx, axis=-2) + X * measure
    D = math.gather(math.gather(A, idx, axis=-1), not_idx, axis=-2)
    R = math.gather(math.gather(A, not_idx, axis=-1), not_idx, axis=-2)

    bM = math.gather(b, idx, axis=-1)
    bR = math.gather(b, not_idx, axis=-1)

    A_post = R - math.matmul(D, math.inv(M), math.transpose(D))
    b_post = bR - math.matvec(D, math.solve(M, bM))
    c_post = (
        c * math.sqrt((-1) ** n / math.det(M)) * math.exp(-0.5 * math.sum(bM * math.solve(M, bM)))
    )

    return A_post, b_post, c_post


def join_Abc(Abc1, Abc2):
    r"""Joins two ``(A,b,c)`` triples into a single ``(A,b,c)`` triple by block addition of the ``A``
    matrices and concatenating the ``b`` vectors.

    Arguments:
        Abc1: the first ``(A,b,c)`` triple
        Abc2: the second ``(A,b,c)`` triple

    Returns:
        The joined ``(A,b,c)`` triple
    """
    A1, b1, c1 = Abc1
    A2, b2, c2 = Abc2
    A12 = math.block_diag(A1, A2)
    b12 = math.concat([b1, b2], axis=-1)
    c12 = math.outer(c1, c2)
    return A12, b12, c12


def reorder_abc(Abc: tuple, order: Sequence[int]):
    r"""
    Reorders the indices of the A matrix and b vector of an (A,b,c) triple.

    Arguments:
        Abc: the ``(A,b,c)`` triple
        order: the new order of the indices

    Returns:
        The reordered ``(A,b,c)`` triple
    """
    A, b, c = Abc
    A = math.gather(math.gather(A, order, axis=-1), order, axis=-2)
    b = math.gather(b, order, axis=-1)
    if len(c.shape) == len(order):
        c = math.transpose(c, order)
    return A, b, c


def contract_two_Abc(
    Abc1: Tuple[ComplexMatrix, ComplexVector, complex],
    Abc2: Tuple[ComplexMatrix, ComplexVector, complex],
    idx1: Sequence[int],
    idx2: Sequence[int],
):
    r"""
    Returns the contraction of two ``(A,b,c)`` triples.

    Arguments:
        Abc1: the first ``(A,b,c)`` triple
        Abc2: the second ``(A,b,c)`` triple
        idx1: the indices of the first ``(A,b,c)`` triple to contract
        idx2: the indices of the second ``(A,b,c)`` triple to contract

    Returns:
        The contracted ``(A,b,c)`` triple
    """
    Abc = join_Abc(Abc1, Abc2)
    return complex_gaussian_integral(
        Abc, idx1, tuple(n + Abc1[0].shape[-1] for n in idx2), measure=-1.0
    )
